#!/usr/bin/env python
"""
The image calibration module contains functions for calibrating optical data from DN to radience and top of atmosphere reflectance and, using coefficients from 6S, surface reflectance.

For obtaining the correct parameters for each function it is recomented that rather than running directly they are called through the Atmospheric and Radiometric Correction of Satellite Imagery (ARCSI) software. 

More details on ARCSI are avaialble from https://remotesensing.info/arcsi
"""

# import the C++ extension into this level
from ._imagecalibration import *

import rsgislib


class ElevLUTFeat(object):
    """
    Create a list of these objects to pass to the rsgislib.imagecalibration.apply6SCoeffElevLUTParam

    :param Elev: is the elevation value.
    :param Coeffs: is the 6S coeffients as a Band6S object.
    """

    def __init__(self, Elev=None, Coeffs=None):
        """
        :param Elev: is the elevation value.
        :param Coeffs: is the 6S coeffients as a Band6S object.
        """
        self.Elev = Elev
        self.Coeffs = Coeffs


class AOTLUTFeat(object):
    """
    Create a list of these objects to pass to the rsgislib.imagecalibration.apply6SCoeffElevAOTLUTParam

    :param AOT: is the AOT value.
    :param Coeffs: is the 6S coeffients as a Band6S object.
    """

    def __init__(self, AOT=None, Coeffs=None):
        """
        :param AOT: is the AOT value.
        :param Coeffs: is the 6S coeffients as a Band6S object.
        """
        self.AOT = AOT
        self.Coeffs = Coeffs


class Band6SCoeff(object):
    """
    Create a list of these objects to provide the Coeffs for ElevLUTFeat and AOTLUTFeat.
    """

    def __init__(
        self,
        band=None,
        aX=None,
        bX=None,
        cX=None,
        DirIrr=None,
        DifIrr=None,
        EnvIrr=None,
    ):
        self.band = band
        self.aX = aX
        self.bX = bX
        self.cX = cX
        self.DirIrr = DirIrr
        self.DifIrr = DifIrr
        self.EnvIrr = EnvIrr


def perform_dos_calc(
    inputFile,
    outputFile,
    gdalformat="KEA",
    nonNegative=True,
    noDataVal=0,
    darkObjReflVal=0,
    darkObjPercentile=0.01,
    copyBandNames=True,
    calcStatsPyd=True,
):
    """
    A command to perform a dark object subtraction (DOS) on an input image.

    :param inputFile: input image to which the DOS method is to be applied. Typically, this image with be in top of atmosphere reflectance (TOA)
    :param outputFile: the output image file
    :param gdalformat: the output image file format (default = KEA)
    :param nonNegative: is a boolean specifying where negative output pixel values will be accepted (Dafualt is True; i.e., no negative values)
    :param noDataVal: is the no data value within the input image file.
    :param darkObjReflVal: is an offset which is applied to all pixel values to make a minimum reflectance value (Default = 0)
    :param darkObjPercentile: is the percentile of the input image used to define the dark object threshold, range is 0 - 1 (Default is 0.01; i.e., 1%).
    :param copyBandNames: is a boolean specifying that the band names of the input image should be copied to the output image file (Default: True)
    :param calcStatsPyd: is a boolean specifying that the image stats and pyramids should be calculated on the output image (Default: True)

    .. code:: python

        import rsgislib.imagecalibration
        rsgislib.imagecalibration.perform_dos_calc("LS5TM_20110701_lat52lon421_r24p204_rad_toa.kea", 'LS5TM_20110701_lat52lon421_r24p204_rad_toa_dos.kea")

    """
    import rsgislib
    import rsgislib.imagecalc
    import rsgislib.imageutils
    import collections

    outDataType = rsgislib.imageutils.get_rsgislib_datatype_from_img(inputFile)

    percentList = rsgislib.imagecalc.calc_band_percentile(
        inputFile, darkObjPercentile, noDataVal
    )

    offsetsList = list()
    OffVal = collections.namedtuple("DOSOffset", ["offset"])
    for val in percentList:
        offsetsList.append(OffVal(offset=val))

    apply_subtract_single_offsets(
        inputFile,
        outputFile,
        gdalformat,
        outDataType,
        nonNegative,
        True,
        noDataVal,
        darkObjReflVal,
        offsetsList,
    )

    if copyBandNames:
        bandNames = rsgislib.imageutils.getBandNames(inputFile)
        rsgislib.imageutils.set_band_names(outputFile, bandNames)

    if calcStatsPyd:
        rsgislib.imageutils.pop_img_stats(
            outputFile, use_no_data=True, no_data_val=noDataVal, calc_pyramids=True
        )


def calc_clear_sky_regions(
    cloudsImg,
    validAreaImg,
    outputClearSkyMask,
    outFormat,
    tmpPath="tmpClearSky",
    deleteTmpFiles=True,
    initClearSkyRegionDist=5000,
    initClearSkyRegionMinSize=3000,
    finalClearSkyRegionDist=1000,
    morphSize=21,
):
    """
    Given a cloud mask, identify the larger extent regions of useful clear-sky regions.

    :param cloudsImg: An image with the input mask of the cloud (pixel == 1) and shadow (pixel == 2)
    :param validAreaImg: A mask of the image data area (1 = valid and 0 = not-valid; i.e., outside of the data area)
    :param outputClearSkyMask: The output mask of the clear sky areas
    :param outFormat: The output image format.
    :param tmpPath: The path for temporay images produced during the processing to be stored (Default: 'tmpClearSky'; Note. all temp files are generated as KEA files).
    :param deleteTmpFiles: Boolean as to whether the intermediate files should be deleted following processing (Default: True - delete files).
    :param initClearSkyRegionDist: The distance in metres from a cloud/shadow object for the initial identification of clear sky regions (Default: 5000)
    :param initClearSkyRegionMinSize: The minimum size (in pixels) of the initial clear sky regions (Default: 3000 pixels)
    :param finalClearSkyRegionDist: The distance in metres from a cloud/shadow object for the final boundaries of the clear sky regions (Default: 1000)
    :param morphSize: the size of the circular morphological operator used to tidy up the result (Default: 21)

    .. code:: python

        import rsgislib.imagecalibration
        cloudsImg = "Outputs/LS8_20160605_lat52lon261_r24p203_clouds.kea"
        validAreaImg = "Outputs/LS8_20160605_lat52lon261_r24p203_valid.kea"
        outputMask = "Outputs/LS8_20160605_lat52lon261_r24p203_openskyvalid.kea"
        tmpPath = "temp"
        rsgislib.imagecalibration.calc_clear_sky_regions(cloudsImg, validAreaImg, outputMask, 'KEA', tmpPath)

    """

    import rsgislib
    import rsgislib.imagecalc
    import rsgislib.imageutils
    import rsgislib.segmentation
    import rsgislib.rastergis
    import rsgislib.imagemorphology
    import os
    from osgeo import gdal
    from rios import rat
    import numpy

    if morphSize % 2 == 0:
        raise rsgislib.RSGISPyException(
            "The size of the morphology operator must be odd."
        )

    baseDataName = os.path.splitext(os.path.basename(cloudsImg))[0]
    tmpCloudsImgDist2Clouds = os.path.join(tmpPath, baseDataName + "_dist2clouds.kea")
    tmpCloudsImgDist2CloudsNoData = os.path.join(
        tmpPath, baseDataName + "_dist2clouds_masked.kea"
    )
    tmpInitClearSkyRegions = os.path.join(tmpPath, baseDataName + "initclearsky.kea")
    tmpInitClearSkyRegionsClumps = os.path.join(
        tmpPath, baseDataName + "initclearskyClumps.kea"
    )
    tmpInitClearSkyRegionsRmSmall = os.path.join(
        tmpPath, baseDataName + "initclearskyClumpsRMSmall.kea"
    )
    tmpInitClearSkyRegionsFinal = os.path.join(
        tmpPath, baseDataName + "initclearskyClumpsFinal.kea"
    )
    tmpClearSkyRegionsFullExtent = os.path.join(
        tmpPath, baseDataName + "clearskyClumpsFullExtent.kea"
    )
    tmpClearSkyRegionsFullExtentClumps = os.path.join(
        tmpPath, baseDataName + "clearskyClumpsFullExtentClumps.kea"
    )
    tmpClearSkyRegionsFullExtentSelectClumps = os.path.join(
        tmpPath, baseDataName + "clearskyClumpsFullExtentSelectClumps.kea"
    )
    tmpClearSkyRegionsFullExtentSelectClumpsOpen = os.path.join(
        tmpPath, baseDataName + "clearskyClumpsFullExtentSelectClumpsOpen.kea"
    )
    tmpClearSkyRegionsFullExtentSelectClumpsOpenClump = os.path.join(
        tmpPath, baseDataName + "clearskyClumpsFullExtentSelectClumpsOpenClump.kea"
    )
    tmpClearSkyRegionsFullExtentSelectClumpsOpenClumpRMSmall = os.path.join(
        tmpPath,
        baseDataName + "clearskyClumpsFullExtentSelectClumpsOpenClumpRMSmall.kea",
    )
    tmpMorphOperator = os.path.join(tmpPath, "CircularMorphOp.gmtxt")

    rsgislib.imagecalc.calc_dist_to_img_vals(
        cloudsImg, tmpCloudsImgDist2Clouds, pxl_vals=[1, 2], out_no_data_val=-9999
    )

    rsgislib.imageutils.mask_img(
        tmpCloudsImgDist2Clouds,
        validAreaImg,
        tmpCloudsImgDist2CloudsNoData,
        "KEA",
        rsgislib.TYPE_32INT,
        -1,
        0,
    )

    rsgislib.imagecalc.image_math(
        tmpCloudsImgDist2CloudsNoData,
        tmpInitClearSkyRegions,
        "b1 > " + str(initClearSkyRegionDist),
        outFormat,
        rsgislib.TYPE_32UINT,
    )

    rsgislib.segmentation.clump(
        tmpInitClearSkyRegions, tmpInitClearSkyRegionsClumps, "KEA", False, 0.0, False
    )

    rsgislib.rastergis.pop_rat_img_stats(tmpInitClearSkyRegionsClumps, True, True)

    rsgislib.segmentation.rm_small_clumps(
        tmpInitClearSkyRegionsClumps,
        tmpInitClearSkyRegionsRmSmall,
        initClearSkyRegionMinSize,
        "KEA",
    )

    rsgislib.segmentation.relabel_clumps(
        tmpInitClearSkyRegionsRmSmall, tmpInitClearSkyRegionsFinal, "KEA", False
    )

    rsgislib.rastergis.pop_rat_img_stats(tmpInitClearSkyRegionsFinal, True, True)

    rsgislib.imagecalc.image_math(
        tmpCloudsImgDist2CloudsNoData,
        tmpClearSkyRegionsFullExtent,
        "b1 > " + str(finalClearSkyRegionDist),
        outFormat,
        rsgislib.TYPE_32UINT,
    )

    rsgislib.segmentation.clump(
        tmpClearSkyRegionsFullExtent,
        tmpClearSkyRegionsFullExtentClumps,
        "KEA",
        False,
        0.0,
        False,
    )

    rsgislib.rastergis.pop_rat_img_stats(tmpClearSkyRegionsFullExtentClumps, True, True)

    rsgislib.rastergis.populate_rat_with_stats(
        tmpInitClearSkyRegionsFinal,
        tmpClearSkyRegionsFullExtentClumps,
        [rsgislib.rastergis.BandAttStats(band=1, max_field="InitRegionInter")],
    )

    ratDataset = gdal.Open(tmpClearSkyRegionsFullExtentClumps, gdal.GA_Update)
    InitRegionInter = rat.readColumn(ratDataset, "InitRegionInter")
    ValidClumps = numpy.zeros_like(InitRegionInter, dtype=numpy.dtype("int"))
    ValidClumps[InitRegionInter > 0] = 1
    rat.writeColumn(ratDataset, "ValidClumps", ValidClumps)
    ratDataset = None

    rsgislib.rastergis.collapse_rat(
        tmpClearSkyRegionsFullExtentClumps,
        "ValidClumps",
        tmpClearSkyRegionsFullExtentSelectClumps,
        "KEA",
        1,
    )

    rsgislib.rastergis.pop_rat_img_stats(
        tmpClearSkyRegionsFullExtentSelectClumps, True, True
    )

    rsgislib.imagemorphology.create_circular_op(
        output_file=tmpMorphOperator, op_size=morphSize
    )

    rsgislib.imagemorphology.image_opening(
        input_img=tmpClearSkyRegionsFullExtentSelectClumps,
        output_img=tmpClearSkyRegionsFullExtentSelectClumpsOpen,
        tmp_img="",
        morph_op_file=tmpMorphOperator,
        use_op_file=True,
        op_size=21,
        gdalformat="KEA",
        datatype=rsgislib.TYPE_32UINT,
    )

    rsgislib.segmentation.clump(
        tmpClearSkyRegionsFullExtentSelectClumpsOpen,
        tmpClearSkyRegionsFullExtentSelectClumpsOpenClump,
        "KEA",
        False,
        0.0,
        False,
    )

    rsgislib.rastergis.pop_rat_img_stats(
        tmpClearSkyRegionsFullExtentSelectClumpsOpenClump, True, True
    )

    rsgislib.segmentation.rm_small_clumps(
        tmpClearSkyRegionsFullExtentSelectClumpsOpenClump,
        tmpClearSkyRegionsFullExtentSelectClumpsOpenClumpRMSmall,
        initClearSkyRegionMinSize,
        "KEA",
    )

    rsgislib.imagecalc.image_math(
        tmpClearSkyRegionsFullExtentSelectClumpsOpenClumpRMSmall,
        outputClearSkyMask,
        "b1>0?1:0",
        outFormat,
        rsgislib.TYPE_8UINT,
    )

    if deleteTmpFiles:
        import rsgislib.tools.filetools

        rsgislib.tools.filetools.delete_file_with_basename(tmpCloudsImgDist2Clouds)
        rsgislib.tools.filetools.delete_file_with_basename(
            tmpCloudsImgDist2CloudsNoData
        )
        rsgislib.tools.filetools.delete_file_with_basename(tmpInitClearSkyRegions)
        rsgislib.tools.filetools.delete_file_with_basename(tmpInitClearSkyRegionsClumps)
        rsgislib.tools.filetools.delete_file_with_basename(
            tmpInitClearSkyRegionsRmSmall
        )
        rsgislib.tools.filetools.delete_file_with_basename(tmpInitClearSkyRegionsFinal)
        rsgislib.tools.filetools.delete_file_with_basename(tmpClearSkyRegionsFullExtent)
        rsgislib.tools.filetools.delete_file_with_basename(
            tmpClearSkyRegionsFullExtentClumps
        )
        rsgislib.tools.filetools.delete_file_with_basename(
            tmpClearSkyRegionsFullExtentSelectClumps
        )
        rsgislib.tools.filetools.delete_file_with_basename(
            tmpClearSkyRegionsFullExtentSelectClumpsOpen
        )
        rsgislib.tools.filetools.delete_file_with_basename(
            tmpClearSkyRegionsFullExtentSelectClumpsOpenClump
        )
        rsgislib.tools.filetools.delete_file_with_basename(
            tmpClearSkyRegionsFullExtentSelectClumpsOpenClumpRMSmall
        )
        rsgislib.tools.filetools.delete_file_with_basename(tmpMorphOperator)


def get_esun_value(radiance, toaRefl, day, month, year, solarZenith):
    """
    Get the ESUN value where a radiance and TOA Reflectance value are known for a pixel.

    :param radiance: input at sensor radiance value.
    :param toaRefl: input the known at sensor (top of atmosphere) reflectance value for the given radiance.
    :param day: input the day of the acquisition.
    :param month: input the month of the acquisition.
    :param year: input the year of the acquisition.
    :param solarZenith: input the solar zenith angle for the acquisition.
    :returns: esun radiance
    """
    import rsgislib.imagecalibration
    import math

    julianDay = rsgislib.imagecalibration.get_julian_day(year, month, day)
    solarDist = rsgislib.imagecalibration.calc_solar_distance(julianDay)
    # pi * L * d2
    step1 = math.pi * radiance * (solarDist * solarDist)
    # step1 / toaRefl
    step2 = step1 / toaRefl
    # step2 / cos(solarZenith)
    esun = step2 / math.cos(math.radians(solarZenith))
    return esun


"""
def createEstimateSREFSurface(inputTOAImg, imgBands, bandRescale, winSize, outImage, gdalformat, dataType, tmpDIR):
    ###
    Estimate SREF surface from TOA input image using the method details in He, K., Sun, J., & Tang, X. (2011). 
    'Single image haze removal using dark channel prior'. IEEE Transactions on Pattern Analysis and Machine Intelligence.
    Method details were also identified with reference to https://www.kaggle.com/ivanl1/haze-removal-using-dark-channel-prior
    
    **WARNING DO NOT USE - THIS FUNCTION IS STILL A WORK IN PROGRESS**
    
    ###
    raise rsgislib.RSGISPyException("This function is not yet ready - needs more thought as how to get a fully working implementation.")
    import rsgislib
    import rsgislib.imagecalc
    import rsgislib.imageutils
    
    import os.path
    import shutil
    
    from rios import applier
    from rios import cuiprogress
    
    import numpy
    
    baseName = os.path.splitext(os.path.basename(inputTOAImg))[0]    
    tmpImgDIR = os.path.join(tmpDIR, baseName)
    
    tmpPresent = True
    if not os.path.exists(tmpImgDIR):
        os.makedirs(tmpImgDIR)
        tmpPresent = False 
    
    ###############################################
    ########## NEEDS TO CHANGE SCALING ############
    ###############################################
    #
    # The scaling used seems to have a significant 
    # effect on the results. Should a histogram or
    # standard deviation approach be adopted?
    #
    ###############################################
    
    # Rescale input TOA 
    rescaledInputImg = os.path.join(tmpImgDIR, baseName+'_rescaled.kea')
    rsgislib.imagecalc.rescale_img_pxl_vals(inputTOAImg, rescaledInputImg, 'KEA', rsgislib.TYPE_32FLOAT, bandRescale, trim2Limits=True)
    
    # Calculate the minium value (across all bands) within the window.
    minValInWinImg = os.path.join(tmpImgDIR, baseName+'_minValInWin.kea')
    minValInWinRefImg = os.path.join(tmpImgDIR, baseName+'_minValInWinRef.kea')
    rsgislib.imagecalc.identifyMinPxlValueInWin(rescaledInputImg, minValInWinImg, minValInWinRefImg, imgBands, winSize, 'KEA', nodataval=0.0, usenodata=True)
    
    ###############################################
    ########## NEEDS TO ADD FILTERING #############
    ###############################################
    #
    # Published paper uses the Soft Matting algorithm 
    # but this uses a lot of memory etc. so prob not
    # suitable.
    #
    # A Guided Filter is an alternative and should be 
    # implmented in RSGISLIb - would generally be useful.
    #
    ###############################################
    
    # Calculate the 99th percentile.
    percentVals = rsgislib.imagecalc.calc_band_percentile(minValInWinImg, 0.99, 0)
    
    # Create mask of regions over 99th percentile
    atmosMaskImg = os.path.join(tmpImgDIR, baseName+'_atmosMask.kea')
    rsgislib.imagecalc.image_math(minValInWinImg, atmosMaskImg, '(b1==0)?0:(b1>='+str(percentVals[0])+')?1:0', 'KEA', rsgislib.TYPE_8UINT)
    
    # Calculate the mean value within the atmospheric light mask
    imgMeanVal = rsgislib.imagecalc.calcImgMeanInMask(rescaledInputImg, atmosMaskImg, 1, imgBands, 1, True)
    
    # Calculate the atmospheric transmission
    # transmission = 1 - omega * imgDark / atomsphericLight
    transImg = os.path.join(tmpImgDIR, baseName+'_transmission.kea')
    rsgislib.imagecalc.image_math(minValInWinImg, transImg, '(b1==0)?0:(1 - (1 * (b1/'+str(imgMeanVal)+')))<0.025?0.025:(1 - (1 * (b1/'+str(imgMeanVal)+')))', 'KEA', rsgislib.TYPE_32FLOAT)
    
    #####################################################################
    # Calculate the scaled SREF pixel values.
    srefScaledImg = os.path.join(tmpImgDIR, baseName+'_scaledSREF.kea')
    
    try:
        import tqdm
        progress_bar = rsgislib.TQDMProgressBar()
    except:
        progress_bar = cuiprogress.GDALProgressBar()
    
    infiles = applier.FilenameAssociations()
    infiles.toaScaledImage = rescaledInputImg
    infiles.transImg = transImg
    outfiles = applier.FilenameAssociations()
    outfiles.outimage = srefScaledImg
    otherargs = applier.OtherInputs()
    otherargs.atmosLight = imgMeanVal
    otherargs.imgBands = imgBands
    otherargs.numpyDT = numpy.float32
    aControls = applier.ApplierControls()
    aControls.progress = progress_bar
    aControls.drivername = 'KEA'
    aControls.omitPyramids = True
    aControls.calcStats = False
    
    def _calcScaleSREFVals(info, inputs, outputs, otherargs):
        # This is an internal rios function 
        outImgShape = [len(otherargs.imgBands), inputs.transImg.shape[1], inputs.transImg.shape[2]]
        outputs.outimage = numpy.zeros(outImgShape, dtype=otherargs.numpyDT)
        for idx in range(len(imgBands)):
            bandIdx = imgBands[idx]-1
            outputs.outimage[idx] = numpy.where(inputs.toaScaledImage[bandIdx] == 0.0, 0.0, (((inputs.toaScaledImage[bandIdx]-otherargs.atmosLight))/inputs.transImg[0])+otherargs.atmosLight)

    applier.apply(_calcScaleSREFVals, infiles, outfiles, otherargs, controls=aControls)
    # End of RIOS Section
    #####################################################################
    
    ###############################################
    ########## NEEDS TO CHANGE SCALING ############
    ###############################################
    #
    # How to reliably rescale the results back to 
    # to the original range. (I guess we need the
    # original data range rather than the editted
    # range from the stretch?)
    #
    ###############################################
    
    # Rescale the SREF image
    bandRescaleSREF = []
    outBand = 1
    for band in imgBands:
        bandRescaleSREF.append(rsgislib.imagecalc.ImageBandRescale(outBand, bandRescale[band-1].outMin, bandRescale[band-1].outMax, bandRescale[band-1].outNoData, 5, (bandRescale[band-1].inMax-bandRescale[band-1].inMin), 0))
        outBand = outBand + 1
        
    rsgislib.imagecalc.rescale_img_pxl_vals(srefScaledImg, outImage, gdalformat, dataType, bandRescaleSREF, trim2Limits=True)
    rsgislib.imageutils.pop_img_stats(outImage, use_no_data=True, no_data_val=0, calc_pyramids=True)
    
    if not tmpPresent:
        shutil.rmtree(tmpImgDIR, ignore_errors=True)
"""



import os

DATA_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data", "elevation")


def test_fft_dem_fusion(tmp_path):
    import rsgislib.elevation
    import rsgislib.imagecalc

    high_res_dem_img = os.path.join(DATA_DIR, "SRTM_aber.kea")
    low_res_dem_img = os.path.join(DATA_DIR, "SRTM_aber_low_res_thresh.kea")
    output_img = os.path.join(tmp_path, "out_fused_dem.kea")
    c_size = 20

    rsgislib.elevation.fft_dem_fusion(
        high_res_dem_img, low_res_dem_img, output_img, c_size
    )

    fused_ref_img = os.path.join(DATA_DIR, "SRTM_aber_fusion.kea")
    img_eq, prop_match = rsgislib.imagecalc.are_imgs_equal(output_img, fused_ref_img)
    assert img_eq


def test_slope(tmp_path):
    import rsgislib.elevation
    import rsgislib.imagecalc

    input_img = os.path.join(DATA_DIR, "SRTM_aber.kea")
    output_img = os.path.join(tmp_path, "out_slope.kea")
    out_angle_unit = "degrees"
    gdalformat = "KEA"
    rsgislib.elevation.slope(input_img, output_img, out_angle_unit, gdalformat)

    slope_ref_img = os.path.join(DATA_DIR, "SRTM_aber_slope_deg.kea")
    img_eq, prop_match = rsgislib.imagecalc.are_imgs_equal(output_img, slope_ref_img)
    assert img_eq


def test_slope_pxl_res_img(tmp_path):
    import rsgislib.elevation
    import rsgislib.imagecalc

    in_dem_img = os.path.join(DATA_DIR, "AberSRTM_4326.kea")
    in_pxl_res_img = os.path.join(DATA_DIR, "AberSRTM_4326_PxlRes.kea")
    output_img = os.path.join(tmp_path, "out_slope.kea")
    out_angle_unit = "degrees"
    gdalformat = "KEA"
    rsgislib.elevation.slope_pxl_res_img(
        in_dem_img, in_pxl_res_img, output_img, out_angle_unit, gdalformat
    )

    slope_ref_img = os.path.join(DATA_DIR, "AberSRTM_4326_slope.kea")
    img_eq, prop_match = rsgislib.imagecalc.are_imgs_equal(output_img, slope_ref_img)
    assert img_eq


def test_aspect(tmp_path):
    import rsgislib.elevation
    import rsgislib.imagecalc

    input_img = os.path.join(DATA_DIR, "SRTM_aber.kea")
    output_img = os.path.join(tmp_path, "out_aspect.kea")
    gdalformat = "KEA"
    rsgislib.elevation.aspect(input_img, output_img, gdalformat)

    aspect_ref_img = os.path.join(DATA_DIR, "SRTM_aber_aspect.kea")
    img_eq, prop_match = rsgislib.imagecalc.are_imgs_equal(output_img, aspect_ref_img)
    print(prop_match)
    assert img_eq


def test_aspect_pxl_res_img(tmp_path):
    import rsgislib.elevation
    import rsgislib.imagecalc

    in_dem_img = os.path.join(DATA_DIR, "AberSRTM_4326.kea")
    in_pxl_res_img = os.path.join(DATA_DIR, "AberSRTM_4326_PxlRes.kea")
    output_img = os.path.join(tmp_path, "out_aspect.kea")
    gdalformat = "KEA"
    rsgislib.elevation.aspect_pxl_res_img(
        in_dem_img, in_pxl_res_img, output_img, gdalformat
    )

    aspect_ref_img = os.path.join(DATA_DIR, "AberSRTM_4326_aspect.kea")
    img_eq, prop_match = rsgislib.imagecalc.are_imgs_equal(output_img, aspect_ref_img)
    assert img_eq


def test_dtm_aspect_median_filter(tmp_path):
    import rsgislib.elevation
    import rsgislib.imagecalc

    input_img = os.path.join(DATA_DIR, "SRTM_aber.kea")
    input_aspect_img = os.path.join(DATA_DIR, "SRTM_aber_aspect.kea")
    output_img = os.path.join(tmp_path, "out_aspect_med_filter.kea")
    aspect_range = 0.5
    win_half_size = 10
    gdalformat = "KEA"
    rsgislib.elevation.dtm_aspect_median_filter(
        input_img, input_aspect_img, output_img, aspect_range, win_half_size, gdalformat
    )

    aspect_med_filter_ref_img = os.path.join(
        DATA_DIR, "SRTM_aber_aspect_med_filter.kea"
    )
    img_eq, prop_match = rsgislib.imagecalc.are_imgs_equal(
        output_img, aspect_med_filter_ref_img
    )
    print(prop_match)
    assert img_eq


def test_fill_dem_soille_gratin_1994(tmp_path):
    import rsgislib.elevation
    import rsgislib.imagecalc

    input_img = os.path.join(DATA_DIR, "SRTM_aber.kea")
    valid_mask_img = os.path.join(DATA_DIR, "SRTM_aber_valid_mask.kea")
    output_img = os.path.join(tmp_path, "out_fill_soille_gratin_1994_tmpout.kea")
    rsgislib.elevation.fill_dem_soille_gratin_1994(
        input_img, valid_mask_img, output_img, "KEA"
    )

    fill_ref_img = os.path.join(DATA_DIR, "SRTM_aber_fill_soille_gratin_1994.kea")
    img_eq, prop_match = rsgislib.imagecalc.are_imgs_equal(output_img, fill_ref_img)
    assert img_eq


def test_plane_fit_detreat_dem(tmp_path):
    import rsgislib.elevation
    import rsgislib.imagecalc

    input_img = os.path.join(DATA_DIR, "SRTM_aber.kea")
    output_img = os.path.join(tmp_path, "out_SRTM_planefitdetreat.kea")
    gdalformat = "KEA"
    window_size = 5
    rsgislib.elevation.plane_fit_detreat_dem(
        input_img, output_img, gdalformat, window_size
    )

    plane_fit_ref_img = os.path.join(DATA_DIR, "SRTM_aber_plane_fit_detreat_dem.kea")
    img_eq, prop_match = rsgislib.imagecalc.are_imgs_equal(
        output_img, plane_fit_ref_img
    )
    print(prop_match)
    assert img_eq


def test_shadow_mask(tmp_path):
    import rsgislib.elevation
    import rsgislib.imagecalc

    input_img = os.path.join(DATA_DIR, "SRTM_aber_subset.kea")
    output_img = os.path.join(tmp_path, "out_SRTM_shadow_mask.kea")
    solar_azimuth = 126.45
    solar_zenith = 35.67
    max_height = 5000
    gdalformat = "KEA"
    rsgislib.elevation.shadow_mask(
        input_img, output_img, solar_azimuth, solar_zenith, max_height, gdalformat
    )

    shadow_mask_ref_img = os.path.join(DATA_DIR, "SRTM_aber_shadow_mask.kea")
    img_eq, prop_match = rsgislib.imagecalc.are_imgs_equal(
        output_img, shadow_mask_ref_img
    )
    assert img_eq


def test_local_incidence_angle(tmp_path):
    import rsgislib.elevation
    import rsgislib.imagecalc

    input_img = os.path.join(DATA_DIR, "SRTM_aber.kea")
    output_img = os.path.join(tmp_path, "out_SRTM_localincangle.kea")
    solar_azimuth = 126.45
    solar_zenith = 35.67
    gdalformat = "KEA"
    rsgislib.elevation.local_incidence_angle(
        input_img, output_img, solar_azimuth, solar_zenith, gdalformat
    )

    inc_angle_ref_img = os.path.join(DATA_DIR, "SRTM_aber_localincangle.kea")
    img_eq, prop_match = rsgislib.imagecalc.are_imgs_equal(
        output_img, inc_angle_ref_img
    )
    assert img_eq


def test_local_existance_angle(tmp_path):
    import rsgislib.elevation
    import rsgislib.imagecalc

    input_img = os.path.join(DATA_DIR, "SRTM_aber.kea")
    output_img = os.path.join(tmp_path, "out_SRTM_localexangle.kea")
    solar_azimuth = 126.45
    solar_zenith = 35.67
    gdalformat = "KEA"
    rsgislib.elevation.local_existance_angle(
        input_img, output_img, solar_azimuth, solar_zenith, gdalformat
    )

    ex_angle_ref_img = os.path.join(DATA_DIR, "SRTM_aber_localexangle.kea")
    img_eq, prop_match = rsgislib.imagecalc.are_imgs_equal(output_img, ex_angle_ref_img)
    assert img_eq


def test_hillshade(tmp_path):
    import rsgislib.elevation
    import rsgislib.imagecalc

    input_img = os.path.join(DATA_DIR, "SRTM_aber.kea")
    output_img = os.path.join(tmp_path, "out_SRTM_hillshade.kea")
    solar_azimuth = 126.45
    solar_zenith = 35.67
    gdalformat = "KEA"
    rsgislib.elevation.hillshade(
        input_img, output_img, solar_azimuth, solar_zenith, gdalformat
    )

    hillshade_ref_img = os.path.join(DATA_DIR, "SRTM_aber_hillshade.kea")
    img_eq, prop_match = rsgislib.imagecalc.are_imgs_equal(
        output_img, hillshade_ref_img
    )
    assert img_eq


def test_hillshade_pxl_res_img(tmp_path):
    import rsgislib.elevation
    import rsgislib.imagecalc

    in_dem_img = os.path.join(DATA_DIR, "AberSRTM_4326.kea")
    in_pxl_res_img = os.path.join(DATA_DIR, "AberSRTM_4326_PxlRes.kea")
    output_img = os.path.join(tmp_path, "out_hillshade.kea")
    solar_azimuth = 126.45
    solar_zenith = 35.67
    gdalformat = "KEA"
    rsgislib.elevation.hillshade_pxl_res_img(
        in_dem_img, in_pxl_res_img, output_img, solar_azimuth, solar_zenith, gdalformat
    )

    hillshade_ref_img = os.path.join(DATA_DIR, "AberSRTM_4326_hillshade.kea")
    img_eq, prop_match = rsgislib.imagecalc.are_imgs_equal(
        output_img, hillshade_ref_img
    )
    assert img_eq



import os
import pytest

MATPLOTLIB_NOT_AVAIL = False
try:
    import matplotlib.pyplot
except ImportError:
    MATPLOTLIB_NOT_AVAIL = True

DATA_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data")


def test_get_gdal_raster_mpl_imshow_basic():
    import rsgislib.tools.plotting

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    rsgislib.tools.plotting.get_gdal_raster_mpl_imshow(input_img, bands=None, bbox=None)


def test_get_gdal_raster_mpl_imshow_bands():
    import rsgislib.tools.plotting

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    rsgislib.tools.plotting.get_gdal_raster_mpl_imshow(
        input_img, bands=[8, 9, 3], bbox=None
    )


def test_get_gdal_raster_mpl_imshow_bands_bbox():
    import rsgislib.tools.plotting
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    sub_bbox = rsgislib.imageutils.get_img_bbox(input_img)
    rsgislib.tools.plotting.get_gdal_raster_mpl_imshow(
        input_img, bands=[8, 9, 3], bbox=sub_bbox
    )


def test_linear_stretch_np_arr_3_bands():
    import rsgislib.tools.plotting

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    img_data, img_coords = rsgislib.tools.plotting.get_gdal_raster_mpl_imshow(
        input_img, bands=[1, 2, 3], bbox=None
    )
    rsgislib.tools.plotting.linear_stretch_np_arr(img_data, no_data_val=0.0)


def test_linear_stretch_np_arr_1_band():
    import rsgislib.tools.plotting

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    img_data, img_coords = rsgislib.tools.plotting.get_gdal_raster_mpl_imshow(
        input_img, bands=[1], bbox=None
    )
    rsgislib.tools.plotting.linear_stretch_np_arr(img_data, no_data_val=0.0)


def test_cumulative_stretch_np_arr_3_bands():
    import rsgislib.tools.plotting

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    img_data, img_coords = rsgislib.tools.plotting.get_gdal_raster_mpl_imshow(
        input_img, bands=[1, 2, 3], bbox=None
    )
    rsgislib.tools.plotting.cumulative_stretch_np_arr(img_data, no_data_val=0.0)


def test_cumulative_stretch_np_arr_1_band():
    import rsgislib.tools.plotting

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    img_data, img_coords = rsgislib.tools.plotting.get_gdal_raster_mpl_imshow(
        input_img, bands=[1], bbox=None
    )
    rsgislib.tools.plotting.cumulative_stretch_np_arr(img_data, no_data_val=0.0)


def test_stdev_stretch_np_arr_3_bands():
    import rsgislib.tools.plotting

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    img_data, img_coords = rsgislib.tools.plotting.get_gdal_raster_mpl_imshow(
        input_img, bands=[1, 2, 3], bbox=None
    )
    rsgislib.tools.plotting.stdev_stretch_np_arr(img_data, no_data_val=0.0)


def test_stdev_stretch_np_arr_1_band():
    import rsgislib.tools.plotting

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    img_data, img_coords = rsgislib.tools.plotting.get_gdal_raster_mpl_imshow(
        input_img, bands=[1], bbox=None
    )
    rsgislib.tools.plotting.stdev_stretch_np_arr(img_data, no_data_val=0.0)


def test_manual_stretch_np_arr_3_bands():
    import rsgislib.tools.plotting

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    img_data, img_coords = rsgislib.tools.plotting.get_gdal_raster_mpl_imshow(
        input_img, bands=[1, 2, 3], bbox=None
    )

    min_max_vals = list()
    min_max_vals.append({"min": 10, "max": 400})
    min_max_vals.append({"min": 22, "max": 300})
    min_max_vals.append({"min": 1, "max": 120})

    rsgislib.tools.plotting.manual_stretch_np_arr(
        img_data, min_max_vals, no_data_val=0.0
    )


def test_manual_stretch_np_arr_1_band():
    import rsgislib.tools.plotting

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    img_data, img_coords = rsgislib.tools.plotting.get_gdal_raster_mpl_imshow(
        input_img, bands=[1], bbox=None
    )
    min_max_vals = {"min": 10, "max": 400}

    rsgislib.tools.plotting.manual_stretch_np_arr(
        img_data, min_max_vals, no_data_val=0.0
    )


def test_get_gdal_thematic_raster_mpl_imshow_basic():
    import rsgislib.tools.plotting

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_cls.kea")
    rsgislib.tools.plotting.get_gdal_thematic_raster_mpl_imshow(input_img)


def test_get_gdal_thematic_raster_mpl_imshow_bbox():
    import rsgislib.tools.plotting
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_cls.kea")
    sub_bbox = rsgislib.imageutils.get_img_bbox(input_img)
    rsgislib.tools.plotting.get_gdal_thematic_raster_mpl_imshow(
        input_img, bbox=sub_bbox
    )


@pytest.mark.skipif(MATPLOTLIB_NOT_AVAIL, reason="matplotlib dependency not available")
def test_get_gdal_thematic_raster_mpl_imshow_patches():
    import rsgislib.tools.plotting

    cls_names_lut = dict()
    cls_names_lut[1] = "Forest"
    cls_names_lut[2] = "Grass"
    cls_names_lut[3] = "Urban"
    cls_names_lut[4] = "Water"

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_cls.kea")
    rsgislib.tools.plotting.get_gdal_thematic_raster_mpl_imshow(
        input_img, out_patches=True, cls_names_lut=cls_names_lut
    )



import os
import pytest

PYSON_DB_NOT_AVAIL = False
try:
    import pysondb
except ImportError:
    PYSON_DB_NOT_AVAIL = True

PYCURL_DB_NOT_AVAIL = False
try:
    import pycurl
except ImportError:
    PYCURL_DB_NOT_AVAIL = True


DATA_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data")
TOOLS_UTILS_DATA_DIR = os.path.join(DATA_DIR, "tools", "utils")

# Using test ftp service provided by: https://dlptest.com/ftp-test/
ftp_url_ref = "ftp.dlptest.com"
ftp_user_ref = "dlpuser"
ftp_pass_ref = "rNrKYTX9g7z3RgJRmxWuGHbeu"


def test_upload_ftp_file():
    import rsgislib.tools.ftptools

    test_file = os.path.join(TOOLS_UTILS_DATA_DIR, "basic_str.txt")
    rsgislib.tools.ftptools.upload_ftp_file(
        test_file,
        ftp_url=ftp_url_ref,
        remote_path="/rsgislib_test",
        ftp_timeout=300,
        ftp_user=ftp_user_ref,
        ftp_pass=ftp_pass_ref,
        print_info=True,
    )


@pytest.mark.skipif(True, reason="Skip test as timeout can be a problem for ci.")
def test_get_ftp_file_listings():
    import rsgislib.tools.ftptools

    test_file = os.path.join(TOOLS_UTILS_DATA_DIR, "basic_str.txt")
    rsgislib.tools.ftptools.upload_ftp_file(
        test_file,
        ftp_url=ftp_url_ref,
        remote_path="/rsgislib_test/test",
        ftp_timeout=300,
        ftp_user=ftp_user_ref,
        ftp_pass=ftp_pass_ref,
        print_info=True,
    )
    ftp_files, non_dirs_lst = rsgislib.tools.ftptools.get_ftp_file_listings(
        ftp_url=ftp_url_ref,
        ftp_path="/",
        ftp_user=ftp_user_ref,
        ftp_pass=ftp_pass_ref,
        mlsd_not_avail=True,
    )

    assert len(non_dirs_lst) > 0


@pytest.mark.skipif(
    True,
    reason="pysondb dependency not available and test currently skipped as timeout can be a problem for ci.",
)
def test_create_file_listings_db(tmp_path):
    import rsgislib.tools.ftptools

    test_file = os.path.join(TOOLS_UTILS_DATA_DIR, "basic_str.txt")
    rsgislib.tools.ftptools.upload_ftp_file(
        test_file,
        ftp_url=ftp_url_ref,
        remote_path="/rsgislib_test/test",
        ftp_timeout=300,
        ftp_user=ftp_user_ref,
        ftp_pass=ftp_pass_ref,
        print_info=True,
    )

    out_db_file = os.path.join(tmp_path, "db_file.json")
    rsgislib.tools.ftptools.create_file_listings_db(
        db_json=out_db_file,
        ftp_url=ftp_url_ref,
        ftp_path="/",
        ftp_user=ftp_user_ref,
        ftp_pass=ftp_pass_ref,
        mlsd_not_avail=True,
    )

    assert os.path.exists(out_db_file)


def test_download_ftp_file(tmp_path):
    import rsgislib.tools.ftptools

    test_file = os.path.join(TOOLS_UTILS_DATA_DIR, "basic_str.txt")
    rsgislib.tools.ftptools.upload_ftp_file(
        test_file,
        ftp_url=ftp_url_ref,
        remote_path="/rsgislib_test/test",
        ftp_timeout=300,
        ftp_user=ftp_user_ref,
        ftp_pass=ftp_pass_ref,
        print_info=True,
    )

    out_file = os.path.join(tmp_path, "test_dnwld.txt")
    rsgislib.tools.ftptools.download_ftp_file(
        ftp_url=ftp_url_ref,
        remote_file="/rsgislib_test/test/basic_str.txt",
        local_file=out_file,
        ftp_timeout=300,
        ftp_user=ftp_user_ref,
        ftp_pass=ftp_pass_ref,
    )

    assert os.path.exists(out_file)


@pytest.mark.skipif(PYCURL_DB_NOT_AVAIL, reason="pycurl dependency not available")
def test_download_curl_ftp_file(tmp_path):
    import rsgislib.tools.ftptools

    test_file = os.path.join(TOOLS_UTILS_DATA_DIR, "basic_str.txt")
    rsgislib.tools.ftptools.upload_ftp_file(
        test_file,
        ftp_url=ftp_url_ref,
        remote_path="/rsgislib_test/test",
        ftp_timeout=300,
        ftp_user=ftp_user_ref,
        ftp_pass=ftp_pass_ref,
        print_info=True,
    )

    out_file = os.path.join(tmp_path, "test_dnwld.txt")
    rsgislib.tools.ftptools.download_curl_ftp_file(
        ftp_url=ftp_url_ref,
        remote_file="/rsgislib_test/test/basic_str.txt",
        local_file=out_file,
        ftp_timeout=300,
        ftp_user=ftp_user_ref,
        ftp_pass=ftp_pass_ref,
    )

    assert os.path.exists(out_file)


# @pytest.mark.skipif(PYSON_DB_NOT_AVAIL, reason="pysondb dependency not available")
@pytest.mark.skipif(True, reason="Skipping from automatically running.")
def test_download_files_use_lst_db_no_struct(tmp_path):
    import rsgislib.tools.ftptools

    test_file = os.path.join(TOOLS_UTILS_DATA_DIR, "basic_str.txt")
    rsgislib.tools.ftptools.upload_ftp_file(
        test_file,
        ftp_url=ftp_url_ref,
        remote_path="/rsgislib_test2",
        ftp_timeout=300,
        ftp_user=ftp_user_ref,
        ftp_pass=ftp_pass_ref,
        print_info=True,
    )
    rsgislib.tools.ftptools.upload_ftp_file(
        test_file,
        ftp_url=ftp_url_ref,
        remote_path="/rsgislib_test2/test1",
        ftp_timeout=300,
        ftp_user=ftp_user_ref,
        ftp_pass=ftp_pass_ref,
        print_info=True,
    )
    rsgislib.tools.ftptools.upload_ftp_file(
        test_file,
        ftp_url=ftp_url_ref,
        remote_path="/rsgislib_test2/test2",
        ftp_timeout=300,
        ftp_user=ftp_user_ref,
        ftp_pass=ftp_pass_ref,
        print_info=True,
    )
    rsgislib.tools.ftptools.upload_ftp_file(
        test_file,
        ftp_url=ftp_url_ref,
        remote_path="/rsgislib_test2/test1/test3",
        ftp_timeout=300,
        ftp_user=ftp_user_ref,
        ftp_pass=ftp_pass_ref,
        print_info=True,
    )
    rsgislib.tools.ftptools.upload_ftp_file(
        test_file,
        ftp_url=ftp_url_ref,
        remote_path="/rsgislib_test2/test2/test4",
        ftp_timeout=300,
        ftp_user=ftp_user_ref,
        ftp_pass=ftp_pass_ref,
        print_info=True,
    )

    db_file = os.path.join(tmp_path, "db_file.json")
    rsgislib.tools.ftptools.create_file_listings_db(
        db_json=db_file,
        ftp_url=ftp_url_ref,
        ftp_path="/rsgislib_test2",
        ftp_user=ftp_user_ref,
        ftp_pass=ftp_pass_ref,
        mlsd_not_avail=True,
    )

    out_dir = os.path.join(tmp_path, "tmp")
    os.makedirs(out_dir)

    rsgislib.tools.ftptools.download_files_use_lst_db(
        db_json=db_file,
        out_dir_path=out_dir,
        ftp_timeout=300,
        ftp_user=ftp_user_ref,
        ftp_pass=ftp_pass_ref,
        create_dir_struct=False,
        use_curl=False,
    )


# @pytest.mark.skipif(PYSON_DB_NOT_AVAIL, reason="pysondb dependency not available")
@pytest.mark.skipif(True, reason="Skipping from automatically running.")
def test_download_files_use_lst_db_struct(tmp_path):
    import rsgislib.tools.ftptools

    test_file = os.path.join(TOOLS_UTILS_DATA_DIR, "basic_str.txt")
    rsgislib.tools.ftptools.upload_ftp_file(
        test_file,
        ftp_url=ftp_url_ref,
        remote_path="/rsgislib_test3",
        ftp_timeout=300,
        ftp_user=ftp_user_ref,
        ftp_pass=ftp_pass_ref,
        print_info=True,
    )
    rsgislib.tools.ftptools.upload_ftp_file(
        test_file,
        ftp_url=ftp_url_ref,
        remote_path="/rsgislib_test3/test1",
        ftp_timeout=300,
        ftp_user=ftp_user_ref,
        ftp_pass=ftp_pass_ref,
        print_info=True,
    )
    rsgislib.tools.ftptools.upload_ftp_file(
        test_file,
        ftp_url=ftp_url_ref,
        remote_path="/rsgislib_test3/test2",
        ftp_timeout=300,
        ftp_user=ftp_user_ref,
        ftp_pass=ftp_pass_ref,
        print_info=True,
    )
    rsgislib.tools.ftptools.upload_ftp_file(
        test_file,
        ftp_url=ftp_url_ref,
        remote_path="/rsgislib_test3/test1/test3",
        ftp_timeout=300,
        ftp_user=ftp_user_ref,
        ftp_pass=ftp_pass_ref,
        print_info=True,
    )
    rsgislib.tools.ftptools.upload_ftp_file(
        test_file,
        ftp_url=ftp_url_ref,
        remote_path="/rsgislib_test3/test2/test4",
        ftp_timeout=300,
        ftp_user=ftp_user_ref,
        ftp_pass=ftp_pass_ref,
        print_info=True,
    )

    db_file = os.path.join(tmp_path, "db_file.json")
    rsgislib.tools.ftptools.create_file_listings_db(
        db_json=db_file,
        ftp_url=ftp_url_ref,
        ftp_path="/rsgislib_test2",
        ftp_user=ftp_user_ref,
        ftp_pass=ftp_pass_ref,
        mlsd_not_avail=True,
    )

    out_dir = os.path.join(tmp_path, "tmp")
    os.makedirs(out_dir)

    rsgislib.tools.ftptools.download_files_use_lst_db(
        db_json=db_file,
        out_dir_path=out_dir,
        ftp_timeout=300,
        ftp_user=ftp_user_ref,
        ftp_pass=ftp_pass_ref,
        create_dir_struct=True,
        use_curl=False,
    )



import time
import os
from typing import Dict, List
import ftplib

import rsgislib
import rsgislib.tools.filetools


def _traverse_mlsd_ftp_dirs(
    ftp_conn: ftplib.FTP, ftp_path: str, ftp_files: Dict, try_n_times: int = 5
) -> List[str]:
    """
    An internal function used within the get_ftp_file_listings function
    to recursively retrieve all the files from an FTP server. This function
    uses the mlsd function which isn't supported by all ftp servers but
    is preferable to use if it is supported.

    Note, this is a recursive function which is used internally. You
    probably want to call get_ftp_file_listings to use this function.

    :param ftp_conn: an ftplib connection object.
    :param ftp_path: The path on the ftp server to start from
    :param ftp_files: of the directory and list of files within that directory
    :param try_n_times: if server connection fails try again (sleeping for
                        5 secs in between) n times for failing.
    :return: list of files
    """
    dirs = list()
    non_dirs = list()
    if ftp_path not in ftp_files:
        ftp_files[ftp_path] = list()

    dir_lst = []
    count = 0
    for i in range(try_n_times):
        if count > try_n_times:
            break
        count = count + 1
        try:
            dir_lst = ftp_conn.mlsd(ftp_path, ["type"])
            break
        except Exception as e:
            print("FTP connection failed but trying again: {0}".format(e))
            time.sleep(5)
            continue
    if count > try_n_times:
        raise rsgislib.RSGISPyException(
            "Tried multiple times which failed to get directory "
            "listing on FTP server so failing."
        )

    for item in dir_lst:
        if item[1]["type"] == "dir":
            c_dir = os.path.join(ftp_path, item[0])
            dirs.append(c_dir)
            if c_dir not in ftp_files:
                ftp_files[c_dir] = list()
        elif not ((item[0] == ".") or (item[0] == "..")):
            c_file = os.path.join(ftp_path, item[0])
            non_dirs.append(c_file)
            ftp_files[ftp_path].append(c_file)

    for subdir in sorted(dirs):
        print(subdir)
        tmp_files_lst = _traverse_mlsd_ftp_dirs(
            ftp_conn, subdir, ftp_files, try_n_times
        )
        non_dirs = non_dirs + tmp_files_lst
    return non_dirs


def _traverse_nlst_ftp_dirs(
    ftp_conn: ftplib.FTP, ftp_path: str, ftp_files: Dict, try_n_times: int = 5
) -> List[str]:
    """
    An internal function used within the get_ftp_file_listings function
    to recursively retrieve all the files from an FTP server. This function
    uses the retrlines function which should be supported by all servers.
    Use this function if _traverse_mlsd_ftp_dirs does not work.

    Note, this is a recursive function which is used internally. You
    probably want to call get_ftp_file_listings to use this function.

    :param ftp_conn: an ftplib connection object.
    :param ftp_path: The path on the ftp server to start from
    :param ftp_files: of the directory and list of files within that directory
    :param try_n_times: if server connection fails try again (sleeping for
                        5 secs in between) n times for failing.
    :return: list of files
    """
    dirs = list()
    non_dirs = list()
    if ftp_path not in ftp_files:
        ftp_files[ftp_path] = list()

    ftp_conn.cwd(ftp_path)

    dir_lst = []
    count = 0
    for i in range(try_n_times):
        if count > try_n_times:
            break
        count = count + 1
        try:
            ftp_conn.retrlines("LIST", dir_lst.append)
            break
        except Exception as e:
            print("FTP connection failed but trying again: {0}".format(e))
            time.sleep(5)
            continue
    if count > try_n_times:
        raise rsgislib.RSGISPyException(
            "Tried multiple times which failed to get directory "
            "listing on FTP server so failing."
        )

    dir_lst = map(str.split, dir_lst)
    for item in dir_lst:
        if item[0][0] == "d":
            c_dir = os.path.join(ftp_path, item[-1])
            dirs.append(c_dir)
            if c_dir not in ftp_files:
                ftp_files[c_dir] = list()
        elif not ((item[-1] == ".") or (item[-1] == "..")):
            c_file = os.path.join(ftp_path, item[-1])
            non_dirs.append(c_file)
            ftp_files[ftp_path].append(c_file)

    for subdir in sorted(dirs):
        print(subdir)
        tmp_files_lst = _traverse_nlst_ftp_dirs(
            ftp_conn, subdir, ftp_files, try_n_times
        )
        non_dirs = non_dirs + tmp_files_lst

    return non_dirs


def get_ftp_file_listings(
    ftp_url: str,
    ftp_path: str,
    ftp_user: str,
    ftp_pass: str,
    ftp_timeout: int = 300,
    try_n_times: int = 5,
    mlsd_not_avail: bool = False,
) -> (Dict[str, str], List[str]):
    """
    Traverse the FTP server directory structure to create a list of all the
    files (full paths).

    Note, it is not good practise to hard code passwords into you
    scripts. It would be recommended that you get the password
    from an environmental variable or config file rather than
    hard code the password.

    :param ftp_url: the url for the ftp server
    :param ftp_path: the path on the ftp server where to start the
                     transverse of the tree.
    :param ftp_user: The username for the ftp server
    :param ftp_pass: the password for the ftp server
    :param ftp_timeout: in seconds (None and system default will be used;
                        system defaults are usual above 300 seconds)
    :param try_n_times: if server connection fails try again (sleeping for 5
                        secs in between) n times for failing.
    :param mlsd_not_avail: Option to support older FTP servers which don't
                           support the mlsd function. Default is False but
                           if you get errors then suggest that you try
                           setting this to True and trying again.
    :return: directory by directory and simple list of files as tuple

    """
    ftp_files = dict()
    print("Opening FTP Connection to {}".format(ftp_url))
    ftp_conn = ftplib.FTP(ftp_url, timeout=ftp_timeout)
    ftp_conn.login(user=ftp_user, passwd=ftp_pass)
    print("Traverse the file system and get a list of paths")
    if mlsd_not_avail:
        non_dirs_lst = _traverse_nlst_ftp_dirs(
            ftp_conn, ftp_path, ftp_files, try_n_times
        )
    else:
        non_dirs_lst = _traverse_mlsd_ftp_dirs(
            ftp_conn, ftp_path, ftp_files, try_n_times
        )
    ftp_conn.quit()
    print("Finished traversing the ftp server file system.")
    return ftp_files, non_dirs_lst


def create_file_listings_db(
    db_json: str,
    ftp_url: str,
    ftp_path: str,
    ftp_user: str,
    ftp_pass: str,
    ftp_timeout: int = 300,
    try_n_times: int = 5,
    mlsd_not_avail: bool = False,
):
    """
    A function which transverses the directory structure on an ftp server
    and saves the result to a JSON database using the pysondb module.

    Note, it is not good practise to hard code passwords into you
    scripts. It would be recommended that you get the password
    from an environmental variable or config file rather than
    hard code the password.

    :param db_json: The file path for the databases JSON file.
    :param ftp_url: the url for the ftp server
    :param ftp_path: the path on the ftp server where to start the
                     transverse of the tree.
    :param ftp_user: The username for the ftp server
    :param ftp_pass: the password for the ftp server
    :param ftp_timeout: in seconds (None and system default will be used;
                        system defaults are usual above 300 seconds)
    :param try_n_times: if server connection fails try again (sleeping for 5
                        secs in between) n times for failing.
    :param mlsd_not_avail: Option to support older FTP servers which don't
                           support the mlsd function. Default is False but
                           if you get errors then suggest that you try
                           setting this to True and trying again.

    """
    import pysondb
    import tqdm

    ftp_files, non_dirs_lst = get_ftp_file_listings(
        ftp_url, ftp_path, ftp_user, ftp_pass, ftp_timeout, try_n_times, mlsd_not_avail
    )

    lst_db = pysondb.getDb(db_json)
    db_data = []
    for c_file in tqdm.tqdm(non_dirs_lst):
        db_data.append(
            {
                "ftp_url": ftp_url,
                "rmt_path": c_file,
                "lcl_path": "",
                "downloaded": False,
            }
        )

    if len(db_data) > 0:
        lst_db.addMany(db_data)


def download_ftp_file(
    ftp_url: str,
    remote_file: str,
    local_file: str,
    ftp_timeout: int = 300,
    ftp_user: str = None,
    ftp_pass: str = None,
    print_info: bool = True,
) -> bool:
    """
    A function to download a file from an FTP server.

    :param ftp_url: The remote URL for the ftp server.
    :param remote_file: the remote path on the ftp server for the file to be downloaded
    :param local_file: the local path to where the file should be downloaded to.
    :param ftp_timeout: the timeout for the download. Default: 300 seconds.
    :param ftp_user: the username, if required, for the ftp server.
    :param ftp_pass: the password, if required, for the ftp server.
    :param print_info: bool for whether info should be printed to the console
                       (default: True)
    :return: boolean as to whether the file was successfully downloaded or not.

    """
    try:
        ftp_conn = ftplib.FTP(ftp_url, timeout=ftp_timeout)
        ftp_conn.login(user=ftp_user, passwd=ftp_pass)

        # Write file in binary mode
        if print_info:
            print(f"Downloading: {remote_file}")
        with open(local_file, "wb") as lcl_file_obj:
            # Command for Downloading the file "RETR remote_path"
            ftp_conn.retrbinary(f"RETR {remote_file}", lcl_file_obj.write)
        if print_info:
            print(f"Finished Downloading: {local_file}")
        ftp_conn.quit()
    except:
        return False

    return True


def download_curl_ftp_file(
    ftp_url: str,
    remote_file: str,
    local_file: str,
    ftp_timeout: int = 300,
    ftp_user: str = None,
    ftp_pass: str = None,
    print_info: bool = True,
) -> bool:
    """
    A function to download a file from an FTP server using curl.

    :param ftp_url: The remote URL for the ftp server.
    :param remote_file: the remote path on the ftp server for the file to be downloaded
    :param local_file: the local path to where the file should be downloaded to.
    :param ftp_timeout: the timeout for the download. Default: 300 seconds.
    :param ftp_user: the username, if required, for the ftp server.
    :param ftp_pass: the password, if required, for the ftp server.
    :param print_info: bool for whether info should be printed to the console
                       (default: True)
    :return: boolean as to whether the file was successfully downloaded or not.

    """
    import pycurl

    full_path_url = ftp_url + remote_file
    try:
        fp = open(local_file, "wb")
        curl = pycurl.Curl()
        curl.setopt(pycurl.URL, full_path_url)
        curl.setopt(pycurl.FOLLOWLOCATION, True)
        if print_info:
            curl.setopt(pycurl.NOPROGRESS, 0)
        else:
            curl.setopt(pycurl.NOPROGRESS, 1)
        curl.setopt(pycurl.FOLLOWLOCATION, 1)
        curl.setopt(pycurl.MAXREDIRS, 5)
        curl.setopt(pycurl.CONNECTTIMEOUT, 50)
        curl.setopt(pycurl.TIMEOUT, ftp_timeout)
        curl.setopt(pycurl.FTP_RESPONSE_TIMEOUT, 600)
        curl.setopt(pycurl.NOSIGNAL, 1)
        if (not ftp_user is None) and (not ftp_pass is None):
            curl.setopt(pycurl.HTTPAUTH, pycurl.HTTPAUTH_ANY)
            curl.setopt(pycurl.USERPWD, ftp_user + ":" + ftp_pass)
        curl.setopt(pycurl.WRITEDATA, fp)
        if print_info:
            print("Starting download of {}".format(full_path_url))
        curl.perform()
        if print_info:
            print(
                "Finished download in {0} of {1} bytes for {2}".format(
                    curl.getinfo(curl.TOTAL_TIME),
                    curl.getinfo(curl.SIZE_DOWNLOAD),
                    full_path_url,
                )
            )
        success = True
    except:
        print(
            "An error occurred when downloading {}.".format(
                os.path.join(ftp_url, remote_file)
            )
        )
        success = False
    return success


def download_files_use_lst_db(
    db_json: str,
    out_dir_path: str,
    ftp_timeout: int = 300,
    ftp_user: str = None,
    ftp_pass: str = None,
    create_dir_struct: bool = False,
    use_curl: bool = False,
):
    """
    A function which uses the pysondb JSON database to download all the files
    recording whether files have been downloaded successful and the output
    path for the file.

    :param db_json: file path for the JSON db file.
    :param out_dir_path: the output path where data should be downloaded to.
    :param ftp_timeout: the timeout for the download. Default: 300 seconds.
    :param ftp_user: the username, if required, for the ftp server.
    :param ftp_pass: the password, if required, for the ftp server.
    :param create_dir_struct: boolean specifying whether the folder structure on the
                              ftp server should be maintained within the out_dir_path
                              (True) or ignored and all the individual files just
                              downloaded into the output directory as a flat
                              structure (False; Default).
    :param use_curl: boolean specifying whether to use CURL to download the files.
                     (Default: False).

    """
    import pysondb
    import tqdm

    lst_db = pysondb.getDb(db_json)

    dwld_files = lst_db.getByQuery({"downloaded": False})

    for dwn_file in tqdm.tqdm(dwld_files):
        basename = os.path.basename(dwn_file["rmt_path"])
        if create_dir_struct:
            dir_path = os.path.dirname(dwn_file["rmt_path"])
            if dir_path[0] == "/":
                dir_path = dir_path[1:]
            local_dir_path = os.path.join(out_dir_path, dir_path)
            if not os.path.exists(local_dir_path):
                os.makedirs(local_dir_path)
            local_path = os.path.join(local_dir_path, basename)
        else:
            local_path = os.path.join(out_dir_path, basename)
        if use_curl:
            dwnlded = download_curl_ftp_file(
                ftp_url=dwn_file["ftp_url"],
                remote_file=dwn_file["rmt_path"],
                local_file=local_path,
                ftp_timeout=ftp_timeout,
                ftp_user=ftp_user,
                ftp_pass=ftp_pass,
                print_info=False,
            )
        else:
            dwnlded = download_ftp_file(
                ftp_url=dwn_file["ftp_url"],
                remote_file=dwn_file["rmt_path"],
                local_file=local_path,
                ftp_timeout=ftp_timeout,
                ftp_user=ftp_user,
                ftp_pass=ftp_pass,
                print_info=False,
            )
        if dwnlded:
            lst_db.updateById(
                dwn_file["id"], {"lcl_path": local_path, "downloaded": True}
            )


def upload_ftp_file(
    local_file: str,
    ftp_url: str,
    remote_path: str,
    ftp_timeout: int = 300,
    ftp_user: str = None,
    ftp_pass: str = None,
    print_info: bool = True,
) -> bool:
    """
    A function for uploading a file to an ftp server.

    :param local_file: the path to the local file to be uploaded.
    :param ftp_url: the url for the ftp server for the upload.
    :param remote_path: the path on the ftp server for the file to be uploaded to.
                        this should not include the file name which will be taken
                        from the local_file.
    :param ftp_timeout: the timeout for the download. Default: 300 seconds.
    :param ftp_user: the username, if required, for the ftp server.
    :param ftp_pass: the password, if required, for the ftp server.
    :param print_info: bool for whether info should be printed to the console
                       (default: True)
    :return: boolean as to whether the upload was successfully completed or otherwise.

    """
    if not os.path.isfile(local_file):
        raise rsgislib.RSGISPyException("Input path must be a file.")

    lcl_file_name = os.path.basename(local_file)

    if print_info:
        print("Opening FTP Connection to {}".format(ftp_url))

    ftp_conn = ftplib.FTP(ftp_url, timeout=ftp_timeout)
    ftp_conn.login(user=ftp_user, passwd=ftp_pass)
    if print_info:
        print("FTP Connection to {} is open".format(ftp_url))

    # Check remote path exists and create if it does not.
    remote_path_parts = rsgislib.tools.filetools.split_path_all(remote_path)
    if len(remote_path_parts) > 0:
        if remote_path_parts[0] == "/":
            remote_path_parts.pop(0)

        for rmt_part in remote_path_parts:
            dir_lst = list()
            ftp_conn.retrlines("LIST", dir_lst.append)
            dir_lst = map(str.split, dir_lst)
            found_part = False
            part_dir = False
            for item in dir_lst:
                if item[-1] == rmt_part:
                    found_part = True
                    if item[0][0] == "d":
                        part_dir = True
                    break

            if found_part and (not part_dir):
                raise rsgislib.RSGISPyException("Part of the path was not a directory.")

            if not found_part:
                ftp_conn.mkd(rmt_part)

            ftp_conn.cwd(rmt_part)

    try:
        if print_info:
            print(f"Uploading {lcl_file_name} to FTP Connection.")
        lcl_file_obj = open(local_file, "rb")
        ftp_conn.storbinary(f"STOR {lcl_file_name}", lcl_file_obj)
        lcl_file_obj.close()
        if print_info:
            print(f"Finished Uploading {lcl_file_name}.")

        ftp_conn.quit()
    except:
        return False
    return True



import os
import pytest
import rsgislib.tools.filetools

DATA_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data")
TOOLS_VIS_DATA_DIR = os.path.join(DATA_DIR, "tools", "visualisation")

gdal2tiles_cmd_avail = rsgislib.tools.filetools.is_cmd_tool_avail(
    "gdal2tiles.py", test_call_cmd=["gdal2tiles.py", "-h"]
)
gdal_translate_cmd_avail = rsgislib.tools.filetools.is_cmd_tool_avail(
    "gdal_translate", test_call_cmd=["gdal_translate", "-h"]
)


@pytest.mark.skipif(
    (not gdal_translate_cmd_avail), reason="gdal_translate command not available"
)
def test_create_kmz_img(tmp_path):
    import rsgislib.tools.visualisation

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    output_file = os.path.join(tmp_path, "out_img.kmz")
    rsgislib.tools.visualisation.create_kmz_img(
        input_img,
        output_file,
        "3,2,1",
        reproj_wgs84=True,
        finite_msk=False,
        tmp_dir=tmp_path,
    )

    assert os.path.exists(output_file)


@pytest.mark.skipif(
    (not gdal2tiles_cmd_avail), reason="gdal2tiles.py command not available"
)
def test_create_webtiles_img_no_stats_msk_tms_true(tmp_path):
    import rsgislib.tools.visualisation

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    output_dir = os.path.join(tmp_path, "out_files")
    os.mkdir(output_dir)

    rsgislib.tools.visualisation.create_webtiles_img_no_stats_msk(
        input_img,
        output_dir,
        "3,2,1",
        zoom_levels="2-10",
        resample="average",
        finite_msk=False,
        tms=True,
    )


@pytest.mark.skipif(
    (not gdal2tiles_cmd_avail), reason="gdal2tiles.py command not available"
)
def test_create_webtiles_img_no_stats_msk_tms_false(tmp_path):
    import rsgislib.tools.visualisation

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    output_dir = os.path.join(tmp_path, "out_files")
    os.mkdir(output_dir)

    rsgislib.tools.visualisation.create_webtiles_img_no_stats_msk(
        input_img,
        output_dir,
        "3,2,1",
        zoom_levels="2-10",
        resample="average",
        finite_msk=False,
        tms=False,
    )


@pytest.mark.skipif(
    (not gdal2tiles_cmd_avail), reason="gdal2tiles.py command not available"
)
def test_create_webtiles_img_no_stats_msk_tms_true_finite(tmp_path):
    import rsgislib.tools.visualisation

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    output_dir = os.path.join(tmp_path, "out_files")
    os.mkdir(output_dir)

    rsgislib.tools.visualisation.create_webtiles_img_no_stats_msk(
        input_img,
        output_dir,
        "3,2,1",
        zoom_levels="2-10",
        resample="average",
        finite_msk=True,
        tms=True,
    )


@pytest.mark.skipif(
    (not gdal2tiles_cmd_avail), reason="gdal2tiles.py command not available"
)
def test_create_webtiles_img_nodata_exp(tmp_path):
    import rsgislib.tools.visualisation

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    output_dir = os.path.join(tmp_path, "out_files")
    os.mkdir(output_dir)
    with pytest.raises(rsgislib.RSGISPyException):
        rsgislib.tools.visualisation.create_webtiles_img(
            input_img,
            "3,2,1",
            output_dir,
            zoom_levels="2-10",
            tmp_dir=tmp_path,
            tms=True,
        )


@pytest.mark.skipif(
    (not gdal2tiles_cmd_avail), reason="gdal2tiles.py command not available"
)
def test_create_webtiles_img_tms_true(tmp_path):
    import rsgislib.tools.visualisation

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    output_dir = os.path.join(tmp_path, "out_files")
    os.mkdir(output_dir)

    rsgislib.tools.visualisation.create_webtiles_img(
        input_img,
        "3,2,1",
        output_dir,
        zoom_levels="2-10",
        tmp_dir=tmp_path,
        tms=True,
        no_data_val=0.0,
    )


@pytest.mark.skipif(
    (not gdal2tiles_cmd_avail), reason="gdal2tiles.py command not available"
)
def test_create_webtiles_img_tms_false(tmp_path):
    import rsgislib.tools.visualisation

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    output_dir = os.path.join(tmp_path, "out_files")
    os.mkdir(output_dir)

    rsgislib.tools.visualisation.create_webtiles_img(
        input_img,
        "3,2,1",
        output_dir,
        zoom_levels="2-10",
        tmp_dir=tmp_path,
        tms=False,
        no_data_val=0.0,
    )


@pytest.mark.skipif(
    (not gdal_translate_cmd_avail), reason="gdal_translate command not available"
)
def test_create_quicklook_imgs_sgl(tmp_path):
    import rsgislib.tools.visualisation

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    output_file = os.path.join(tmp_path, "out_img.jpg")
    rsgislib.tools.visualisation.create_quicklook_imgs(
        input_img,
        "3,2,1",
        output_file,
        output_img_sizes=250,
        tmp_dir=tmp_path,
        no_data_val=0.0,
    )

    assert os.path.exists(output_file)


@pytest.mark.skipif(
    (not gdal_translate_cmd_avail), reason="gdal_translate command not available"
)
def test_create_quicklook_imgs_multi(tmp_path):
    import rsgislib.tools.visualisation

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    output_files = [
        os.path.join(tmp_path, "out_img_1.jpg"),
        os.path.join(tmp_path, "out_img_2.jpg"),
    ]
    rsgislib.tools.visualisation.create_quicklook_imgs(
        input_img,
        "3,2,1",
        output_files,
        output_img_sizes=[250, 500],
        tmp_dir=tmp_path,
        no_data_val=0.0,
    )

    assert os.path.exists(output_files[0]) and os.path.exists(output_files[1])


def test_create_mbtile_file(tmp_path):
    import rsgislib.tools.visualisation

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    output_file = os.path.join(tmp_path, "out_img.mbtiles")

    rsgislib.tools.visualisation.create_mbtile_file(
        input_img,
        "3,2,1",
        output_file,
        scale_input_img=50,
        tmp_dir=tmp_path,
        tile_format="PNG",
        no_data_val=0.0,
    )

    assert os.path.exists(output_file)


@pytest.mark.skipif(
    (not gdal_translate_cmd_avail) or (not gdal2tiles_cmd_avail),
    reason="gdal_translate or gdal2tiles.py command not available",
)
def test_create_webtiles_vis_gtiff_img(tmp_path):
    import rsgislib.tools.visualisation

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    output_dir = os.path.join(tmp_path, "out_files")
    output_file = os.path.join(tmp_path, "out_img.tif")

    rsgislib.tools.visualisation.create_webtiles_vis_gtiff_img(
        input_img,
        "3,2,1",
        output_dir,
        output_file,
        tmp_dir=tmp_path,
        webview=True,
        scale=0,
        tms=True,
        no_data_val=0.0,
    )

    assert os.path.exists(output_file)


@pytest.mark.skipif(
    (not gdal_translate_cmd_avail), reason="gdal_translate command not available"
)
def test_create_quicklook_overview_imgs_sgl(tmp_path):
    import rsgislib.tools.visualisation

    input_imgs = [
        os.path.join(DATA_DIR, "sen2_20210527_aber.kea"),
        os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea"),
    ]
    output_file = os.path.join(tmp_path, "out_img.jpg")

    rsgislib.tools.visualisation.create_quicklook_overview_imgs(
        input_imgs,
        "3,2,1",
        tmp_path,
        output_file,
        output_img_sizes=250,
        no_data_val=0.0,
    )

    assert os.path.exists(output_file)


@pytest.mark.skipif(
    (not gdal_translate_cmd_avail), reason="gdal_translate command not available"
)
def test_create_quicklook_overview_imgs_multi(tmp_path):
    import rsgislib.tools.visualisation

    input_imgs = [
        os.path.join(DATA_DIR, "sen2_20210527_aber.kea"),
        os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea"),
    ]
    output_files = [
        os.path.join(tmp_path, "out_img_1.jpg"),
        os.path.join(tmp_path, "out_img_2.jpg"),
    ]

    rsgislib.tools.visualisation.create_quicklook_overview_imgs(
        input_imgs,
        "3,2,1",
        tmp_path,
        output_files,
        output_img_sizes=[250, 500],
        no_data_val=0.0,
    )

    assert os.path.exists(output_files[0]) and os.path.exists(output_files[1])


def test_burn_in_binary_msk(tmp_path):
    import rsgislib.tools.visualisation

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    msk_img = os.path.join(DATA_DIR, "sen2_20210527_aber_vldmsk.kea")
    output_img = os.path.join(tmp_path, "out_img.kea")

    rsgislib.tools.visualisation.burn_in_binary_msk(
        input_img, msk_img, output_img, "KEA", msk_colour=None
    )

    assert os.path.exists(output_img)


def test_create_quicklook_overview_imgs_vec_overlay_sgl(tmp_path):
    import rsgislib.tools.visualisation

    input_imgs = [os.path.join(DATA_DIR, "sen2_20210527_aber.kea")]
    vec_overlay_file = os.path.join(DATA_DIR, "aber_osgb_single_poly.geojson")
    vec_overlay_lyr = "aber_osgb_single_poly"
    output_file = os.path.join(tmp_path, "out_img.jpg")

    rsgislib.tools.visualisation.create_quicklook_overview_imgs_vec_overlay(
        input_imgs,
        "3,2,1",
        tmp_path,
        vec_overlay_file,
        vec_overlay_lyr,
        output_file,
        output_img_sizes=250,
        gdalformat="JPEG",
        no_data_val=0.0,
    )

    assert os.path.exists(output_file)


@pytest.mark.skipif(
    (not gdal_translate_cmd_avail), reason="gdal_translate command not available"
)
def test_create_quicklook_overview_imgs_vec_overlay_sgl_clr(tmp_path):
    import rsgislib.tools.visualisation

    input_imgs = [os.path.join(DATA_DIR, "sen2_20210527_aber.kea")]
    vec_overlay_file = os.path.join(DATA_DIR, "aber_osgb_single_poly.geojson")
    vec_overlay_lyr = "aber_osgb_single_poly"
    output_file = os.path.join(tmp_path, "out_img.jpg")

    rsgislib.tools.visualisation.create_quicklook_overview_imgs_vec_overlay(
        input_imgs,
        "3,2,1",
        tmp_path,
        vec_overlay_file,
        vec_overlay_lyr,
        output_file,
        output_img_sizes=250,
        gdalformat="JPEG",
        overlay_clrs=[255, 0, 0],
        no_data_val=0.0,
    )

    assert os.path.exists(output_file)


@pytest.mark.skipif(
    (not gdal_translate_cmd_avail), reason="gdal_translate command not available"
)
def test_create_quicklook_overview_imgs_vec_overlay_multi(tmp_path):
    import rsgislib.tools.visualisation

    input_imgs = [
        os.path.join(DATA_DIR, "sen2_20210527_aber.kea"),
        os.path.join(DATA_DIR, "sen2_20210527_aber.kea"),
    ]
    vec_overlay_file = os.path.join(DATA_DIR, "aber_osgb_single_poly.geojson")
    vec_overlay_lyr = "aber_osgb_single_poly"
    output_file = os.path.join(tmp_path, "out_img.jpg")

    rsgislib.tools.visualisation.create_quicklook_overview_imgs_vec_overlay(
        input_imgs,
        "3,2,1",
        tmp_path,
        vec_overlay_file,
        vec_overlay_lyr,
        output_file,
        output_img_sizes=250,
        gdalformat="JPEG",
        no_data_val=0.0,
    )

    assert os.path.exists(output_file)


@pytest.mark.skipif(
    (not gdal_translate_cmd_avail), reason="gdal_translate command not available"
)
def test_create_quicklook_overview_imgs_vec_overlay_multi_clr(tmp_path):
    import rsgislib.tools.visualisation

    input_imgs = [
        os.path.join(DATA_DIR, "sen2_20210527_aber.kea"),
        os.path.join(DATA_DIR, "sen2_20210527_aber.kea"),
    ]
    vec_overlay_file = os.path.join(DATA_DIR, "aber_osgb_single_poly.geojson")
    vec_overlay_lyr = "aber_osgb_single_poly"
    output_file = os.path.join(tmp_path, "out_img.jpg")

    rsgislib.tools.visualisation.create_quicklook_overview_imgs_vec_overlay(
        input_imgs,
        "3,2,1",
        tmp_path,
        vec_overlay_file,
        vec_overlay_lyr,
        output_file,
        output_img_sizes=250,
        gdalformat="JPEG",
        overlay_clrs=[255, 0, 0],
        no_data_val=0.0,
    )

    assert os.path.exists(output_file)


@pytest.mark.skipif(
    (not gdal_translate_cmd_avail), reason="gdal_translate command not available"
)
def test_create_quicklook_overview_imgs_vec_overlay_multi_clr_multi_out(tmp_path):
    import rsgislib.tools.visualisation

    input_imgs = [
        os.path.join(DATA_DIR, "sen2_20210527_aber.kea"),
        os.path.join(DATA_DIR, "sen2_20210527_aber.kea"),
    ]
    vec_overlay_file = os.path.join(DATA_DIR, "aber_osgb_single_poly.geojson")
    vec_overlay_lyr = "aber_osgb_single_poly"
    output_files = [
        os.path.join(tmp_path, "out_img_0.jpg"),
        os.path.join(tmp_path, "out_img_1.jpg"),
    ]

    rsgislib.tools.visualisation.create_quicklook_overview_imgs_vec_overlay(
        input_imgs,
        "3,2,1",
        tmp_path,
        vec_overlay_file,
        vec_overlay_lyr,
        output_files,
        output_img_sizes=[250, 500],
        gdalformat="JPEG",
        overlay_clrs=[255, 0, 0],
        no_data_val=0.0,
    )

    assert os.path.exists(output_files[0]) and os.path.exists(output_files[1])


@pytest.mark.skipif(
    (not gdal_translate_cmd_avail), reason="gdal_translate command not available"
)
def test_create_visual_overview_imgs_vec_extent(tmp_path):
    import rsgislib.tools.visualisation

    input_imgs = [os.path.join(DATA_DIR, "sen2_20210527_aber.kea")]
    output_file = os.path.join(tmp_path, "out_img.tif")
    stretch_file = os.path.join(tmp_path, "strch_info.txt")
    rsgislib.tools.visualisation.create_visual_overview_imgs_vec_extent(
        input_imgs,
        "3,2,1",
        tmp_dir=tmp_path,
        vec_extent_file=None,
        vec_extent_lyr=None,
        output_imgs=output_file,
        output_img_sizes=500,
        gdalformat="GTIFF",
        scale_axis="auto",
        stretch_file=stretch_file,
        export_stretch_file=True,
        no_data_val=0.0,
    )

    assert os.path.exists(output_file) and os.path.exists(stretch_file)


@pytest.mark.skipif(
    (not gdal_translate_cmd_avail), reason="gdal_translate command not available"
)
def test_create_visual_overview_imgs_vec_extent_multi_out(tmp_path):
    import rsgislib.tools.visualisation

    input_imgs = [os.path.join(DATA_DIR, "sen2_20210527_aber.kea")]
    output_files = [
        os.path.join(tmp_path, "out_img_0.tif"),
        os.path.join(tmp_path, "out_img_1.tif"),
    ]
    stretch_file = os.path.join(tmp_path, "strch_info.txt")

    rsgislib.tools.visualisation.create_visual_overview_imgs_vec_extent(
        input_imgs,
        "3,2,1",
        tmp_dir=tmp_path,
        vec_extent_file=None,
        vec_extent_lyr=None,
        output_imgs=output_files,
        output_img_sizes=[250, 500],
        gdalformat="GTIFF",
        scale_axis="auto",
        stretch_file=stretch_file,
        export_stretch_file=True,
        no_data_val=0.0,
    )

    assert (
        os.path.exists(output_files[0])
        and os.path.exists(output_files[1])
        and os.path.exists(stretch_file)
    )


@pytest.mark.skipif(
    (not gdal_translate_cmd_avail), reason="gdal_translate command not available"
)
def test_create_visual_overview_imgs_vec_extent_vec_ext(tmp_path):
    import rsgislib.tools.visualisation

    input_imgs = [os.path.join(DATA_DIR, "sen2_20210527_aber.kea")]
    output_file = os.path.join(tmp_path, "out_img.tif")
    stretch_file = os.path.join(tmp_path, "strch_info.txt")
    vec_extent_file = os.path.join(TOOLS_VIS_DATA_DIR, "wales_bbox.geojson")
    vec_extent_lyr = "wales_bbox"

    rsgislib.tools.visualisation.create_visual_overview_imgs_vec_extent(
        input_imgs,
        "3,2,1",
        tmp_dir=tmp_path,
        vec_extent_file=vec_extent_file,
        vec_extent_lyr=vec_extent_lyr,
        output_imgs=output_file,
        output_img_sizes=500,
        gdalformat="GTIFF",
        scale_axis="auto",
        stretch_file=stretch_file,
        export_stretch_file=True,
        no_data_val=0.0,
    )

    assert os.path.exists(output_file) and os.path.exists(stretch_file)


@pytest.mark.skipif(
    (not gdal_translate_cmd_avail), reason="gdal_translate command not available"
)
def test_create_visual_overview_imgs_vec_extent_multi_out_vec_ext(tmp_path):
    import rsgislib.tools.visualisation

    input_imgs = [os.path.join(DATA_DIR, "sen2_20210527_aber.kea")]
    output_files = [
        os.path.join(tmp_path, "out_img_0.tif"),
        os.path.join(tmp_path, "out_img_1.tif"),
    ]
    stretch_file = os.path.join(tmp_path, "strch_info.txt")
    vec_extent_file = os.path.join(TOOLS_VIS_DATA_DIR, "wales_bbox.geojson")
    vec_extent_lyr = "wales_bbox"

    rsgislib.tools.visualisation.create_visual_overview_imgs_vec_extent(
        input_imgs,
        "3,2,1",
        tmp_dir=tmp_path,
        vec_extent_file=vec_extent_file,
        vec_extent_lyr=vec_extent_lyr,
        output_imgs=output_files,
        output_img_sizes=[250, 500],
        gdalformat="GTIFF",
        scale_axis="auto",
        stretch_file=stretch_file,
        export_stretch_file=True,
        no_data_val=0.0,
    )

    assert (
        os.path.exists(output_files[0])
        and os.path.exists(output_files[1])
        and os.path.exists(stretch_file)
    )


def test_overlay_vec_on_img(tmp_path):
    import rsgislib.tools.visualisation

    input_img = os.path.join(
        TOOLS_VIS_DATA_DIR, "sen2_20210527_aber_subset_b123_stch.kea"
    )
    output_img = os.path.join(tmp_path, "out_img.png")
    vec_overlay_file = os.path.join(DATA_DIR, "aber_osgb_single_poly.geojson")
    vec_overlay_lyr = "aber_osgb_single_poly"

    rsgislib.tools.visualisation.overlay_vec_on_img(
        input_img,
        output_img,
        vec_overlay_file,
        vec_overlay_lyr,
        tmp_dir=tmp_path,
        gdalformat="PNG",
        overlay_clrs=[0, 255, 0],
    )

    assert os.path.exists(output_img)



#! /usr/bin/env python
############################################################################
#  tiledsegsingle.py
#
#  Copyright 2016 RSGISLib.
#
#  RSGISLib: 'The remote sensing and GIS Software Library'
#
#  RSGISLib is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  RSGISLib is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with RSGISLib.  If not, see <http://www.gnu.org/licenses/>.
#
#
# Purpose:  Provide functionality to perform the shepherd segmentation
#           algorithm using a tiled implementation but only with a single
#           thread of execution (i.e., it will not used mulitple cores/processors)
#
# Author: Pete Bunting
# Email: petebunting@mac.com
# Date: 4/3/2016
# Version: 1.1
#
# History:
# Version 1.0 - Created.
#
###########################################################################

import glob
import os
import json
import shutil

from osgeo import gdal
from rios import rat

import rsgislib
from rsgislib.imageutils import tilingutils
from rsgislib.segmentation import shepherdseg
from rsgislib import rastergis
from rsgislib import imageutils
from rsgislib import segmentation


class RSGISTiledShepherdSegmentationSingleThread(object):
    """
    A class for running the tiled version of the Shepherd et al (2019) segmentation algorithm.
    This can process larger images than the single scene version with a smaller
    memory footprint.

    Shepherd, J. D., Bunting, P., & Dymond, J. R. (2019). Operational Large-Scale Segmentation of Imagery Based on Iterative Elimination. Remote Sensing, 11(6), 658. http://doi.org/10.3390/rs11060658

    This version can only used a single thread for execution.

    It is not intended that this class will be directly used. Please use the
    function perform_tiled_segmentation to call this functionality.

    """

    def findSegStatsFiles(self, tileImg, segStatsInfo):
        gdalDS = gdal.Open(tileImg, gdal.GA_ReadOnly)
        geotransform = gdalDS.GetGeoTransform()
        if not geotransform is None:
            xMin = geotransform[0]
            yMax = geotransform[3]

            xRes = geotransform[1]
            yRes = geotransform[5]

            width = gdalDS.RasterXSize * xRes
            if yRes < 0:
                yRes = yRes * (-1)
            height = gdalDS.RasterYSize * yRes
            xMax = xMin + width
            yMin = xMax - height

            xCen = xMin + (width / 2)
            yCen = yMin + (height / 2)
        gdalDS = None

        first = True
        minDist = 0.0
        minKCenFile = ""
        minStchStatsFile = ""
        for tileName in segStatsInfo:
            tileXCen = segStatsInfo[tileName]["CENTRE_PT"]["X"]
            tileYCen = segStatsInfo[tileName]["CENTRE_PT"]["Y"]
            dist = ((tileXCen - xCen) * (tileXCen - xCen)) + (
                (tileYCen - yCen) * (tileYCen - yCen)
            )
            if first:
                minKCenFile = segStatsInfo[tileName]["KCENTRES"]
                minStchStatsFile = segStatsInfo[tileName]["STRETCHSTATS"]
                minDist = dist
                first = False
            elif dist < minDist:
                minKCenFile = segStatsInfo[tileName]["KCENTRES"]
                minStchStatsFile = segStatsInfo[tileName]["STRETCHSTATS"]
                minDist = dist

        return minKCenFile, minStchStatsFile

    def performStage1Tiling(
        self,
        inputImage,
        tileShp,
        tilesRat,
        tilesBase,
        tilesMetaDIR,
        tilesImgDIR,
        tmpDIR,
        width,
        height,
        validDataThreshold,
    ):
        tilingutils.createMinDataTiles(
            inputImage,
            tileShp,
            tilesRat,
            width,
            height,
            validDataThreshold,
            None,
            False,
            True,
            tmpDIR,
        )
        tilingutils.createTileMaskImagesFromClumps(
            tilesRat, tilesBase, tilesMetaDIR, "KEA"
        )
        dataType = rsgislib.imageutils.get_rsgislib_datatype_from_img(inputImage)
        tilingutils.createTilesFromMasks(
            inputImage, tilesBase, tilesMetaDIR, tilesImgDIR, dataType, "KEA"
        )

    def performStage1TilesSegmentation(
        self,
        tilesImgDIR,
        stage1TilesSegsDIR,
        tmpDIR,
        tilesBase,
        tileSegInfoJSON,
        strchStatsBase,
        kCentresBase,
        numClustersVal,
        minPxlsVal,
        distThresVal,
        bandsVal,
        samplingVal,
        kmMaxIterVal,
    ):
        imgTiles = glob.glob(os.path.join(tilesImgDIR, tilesBase + "*.kea"))
        tileStatsFiles = dict()
        for imgTile in imgTiles:
            baseName = os.path.splitext(os.path.basename(imgTile))[0]
            tileID = baseName.split("_")[-1]
            clumpsFile = os.path.join(stage1TilesSegsDIR, baseName + "_segs.kea")
            tmpStatsJSON = os.path.join(tilesImgDIR, baseName + "_segstats.json")
            strchStatsOutFile = strchStatsBase + "_" + tileID + ".txt"
            kCentresOutFile = kCentresBase + "_" + tileID
            print(clumpsFile)
            shepherdseg.runShepherdSegmentation(
                imgTile,
                clumpsFile,
                outputMeanImg=None,
                tmpath=os.path.join(tmpDIR, tileID + "_segstemp"),
                gdalformat="KEA",
                noStats=False,
                noStretch=False,
                noDelete=False,
                numClusters=numClustersVal,
                minPxls=minPxlsVal,
                distThres=distThresVal,
                bands=bandsVal,
                sampling=samplingVal,
                kmMaxIter=kmMaxIterVal,
                processInMem=False,
                saveProcessStats=True,
                imgStretchStats=strchStatsOutFile,
                kMeansCentres=kCentresOutFile,
                imgStatsJSONFile=tmpStatsJSON,
            )

            with open(tmpStatsJSON, "r") as f:
                jsonStrData = f.read()
            segStatsInfo = json.loads(jsonStrData)
            tileStatsFiles[baseName] = segStatsInfo
            rsgislib.tools.filetools.delete_file_silent(tmpStatsJSON)

        with open(tileSegInfoJSON, "w") as outfile:
            json.dump(
                tileStatsFiles,
                outfile,
                sort_keys=True,
                indent=4,
                separators=(",", ": "),
                ensure_ascii=False,
            )

    def defineStage1Boundaries(self, tilesImgDIR, stage1TilesSegBordersDIR, tilesBase):
        segTiles = glob.glob(os.path.join(tilesImgDIR, tilesBase + "*_segs.kea"))
        for segTile in segTiles:
            baseName = os.path.splitext(os.path.basename(segTile))[0]
            borderMaskFile = os.path.join(
                stage1TilesSegBordersDIR, baseName + "_segsborder.kea"
            )
            rastergis.defineBorderClumps(segTile, "BoundaryClumps")
            rastergis.export_col_to_gdal_img(
                segTile, borderMaskFile, "KEA", rsgislib.TYPE_8UINT, "BoundaryClumps"
            )

    def mergeStage1TilesToOutput(
        self,
        inputImage,
        tilesSegsDIR,
        tilesSegsBordersDIR,
        tilesBase,
        clumpsImage,
        bordersImage,
    ):
        segTiles = glob.glob(os.path.join(tilesSegsDIR, tilesBase + "*_segs.kea"))
        imageutils.create_copy_img(
            inputImage, clumpsImage, 1, 0, "KEA", rsgislib.TYPE_32UINT
        )
        segmentation.merge_clump_images(segTiles, clumpsImage)
        rastergis.pop_rat_img_stats(clumpsImage, True, True)

        tileBorders = glob.glob(
            os.path.join(tilesSegsBordersDIR, tilesBase + "*_segsborder.kea")
        )
        imageutils.create_copy_img(
            inputImage, bordersImage, 1, 0, "KEA", rsgislib.TYPE_8UINT
        )
        imageutils.includeImages(bordersImage, tileBorders)
        rastergis.pop_rat_img_stats(bordersImage, True, True)

    def performStage2Tiling(
        self,
        inputImage,
        tileShp,
        tilesRat,
        tilesBase,
        tilesMetaDIR,
        tilesImgDIR,
        tmpDIR,
        width,
        height,
        validDataThreshold,
        bordersImage,
    ):
        tilingutils.createMinDataTiles(
            inputImage,
            tileShp,
            tilesRat,
            width,
            height,
            validDataThreshold,
            bordersImage,
            True,
            True,
            tmpDIR,
        )
        tilingutils.createTileMaskImagesFromClumps(
            tilesRat, tilesBase, tilesMetaDIR, "KEA"
        )
        dataType = rsgislib.imageutils.get_rsgislib_datatype_from_img(inputImage)
        tilingutils.createTilesFromMasks(
            inputImage, tilesBase, tilesMetaDIR, tilesImgDIR, dataType, "KEA"
        )

    def performStage2TilesSegmentation(
        self,
        tilesImgDIR,
        tilesMaskedDIR,
        tilesSegsDIR,
        tilesSegBordersDIR,
        tmpDIR,
        tilesBase,
        s1BordersImage,
        segStatsInfo,
        minPxlsVal,
        distThresVal,
        bandsVal,
    ):
        imgTiles = glob.glob(os.path.join(tilesImgDIR, tilesBase + "*.kea"))
        for imgTile in imgTiles:
            baseName = os.path.splitext(os.path.basename(imgTile))[0]
            maskedFile = os.path.join(tilesMaskedDIR, baseName + "_masked.kea")
            dataType = rsgislib.imageutils.get_rsgislib_datatype_from_img(imgTile)
            imageutils.mask_img(
                imgTile, s1BordersImage, maskedFile, "KEA", dataType, 0, 0
            )

        imgTiles = glob.glob(os.path.join(tilesMaskedDIR, tilesBase + "*_masked.kea"))
        for imgTile in imgTiles:
            baseName = os.path.splitext(os.path.basename(imgTile))[0]
            clumpsFile = os.path.join(tilesSegsDIR, baseName + "_segs.kea")
            kMeansCentres, imgStretchStats = self.findSegStatsFiles(
                imgTile, segStatsInfo
            )
            shepherdseg.runShepherdSegmentationPreCalcdStats(
                imgTile,
                clumpsFile,
                kMeansCentres,
                imgStretchStats,
                outputMeanImg=None,
                tmpath=os.path.join(tmpDIR, baseName + "_segstemp"),
                gdalformat="KEA",
                noStats=False,
                noStretch=False,
                noDelete=False,
                minPxls=minPxlsVal,
                distThres=distThresVal,
                bands=bandsVal,
                processInMem=False,
            )

        segTiles = glob.glob(os.path.join(tilesSegsDIR, tilesBase + "*_segs.kea"))
        for segTile in segTiles:
            baseName = os.path.splitext(os.path.basename(segTile))[0]
            borderMaskFile = os.path.join(
                tilesSegBordersDIR, baseName + "_segsborder.kea"
            )
            rastergis.defineBorderClumps(segTile, "BoundaryClumps")
            rastergis.export_col_to_gdal_img(
                segTile, borderMaskFile, "KEA", rsgislib.TYPE_8UINT, "BoundaryClumps"
            )

    def mergeStage2TilesToOutput(
        self, clumpsImage, tilesSegsDIR, tilesSegBordersDIR, tilesBase, s2BordersImage
    ):
        segTiles = glob.glob(os.path.join(tilesSegsDIR, tilesBase + "*_segs.kea"))
        segmentation.merge_clump_images(segTiles, clumpsImage)
        rastergis.pop_rat_img_stats(clumpsImage, True, True)

        tileBorders = glob.glob(
            os.path.join(tilesSegBordersDIR, tilesBase + "*_segsborder.kea")
        )
        imageutils.create_copy_img(
            clumpsImage, s2BordersImage, 1, 0, "KEA", rsgislib.TYPE_8UINT
        )
        imageutils.includeImages(s2BordersImage, tileBorders)

    def createStage3ImageSubsets(
        self,
        inputImage,
        s2BordersImage,
        s3BordersClumps,
        subsetImgsDIR,
        subsetImgsMaskedDIR,
        subImgBaseName,
        minSize,
    ):
        segmentation.clump(s2BordersImage, s3BordersClumps, "KEA", True, 0)
        rastergis.pop_rat_img_stats(s3BordersClumps, True, True)

        rastergis.spatialExtent(
            s3BordersClumps,
            "minXX",
            "minXY",
            "maxXX",
            "maxXY",
            "minYX",
            "minYY",
            "maxYX",
            "maxYY",
        )

        dataType = rsgislib.imageutils.get_rsgislib_datatype_from_img(inputImage)

        ratDS = gdal.Open(s3BordersClumps, gdal.GA_Update)
        minX = rat.readColumn(ratDS, "minXX")
        maxX = rat.readColumn(ratDS, "maxXX")
        minY = rat.readColumn(ratDS, "minYY")
        maxY = rat.readColumn(ratDS, "maxYY")
        Histogram = rat.readColumn(ratDS, "Histogram")
        for i in range(minX.shape[0]):
            if i > 0:
                subImage = os.path.join(subsetImgsDIR, subImgBaseName + str(i) + ".kea")
                # print( "[" + str(minX[i]) + ", " + str(maxX[i]) + "][" + str(minY[i]) + ", " + str(maxY[i]) + "]" )
                imageutils.subsetbbox(
                    inputImage,
                    subImage,
                    "KEA",
                    dataType,
                    minX[i],
                    maxX[i],
                    minY[i],
                    maxY[i],
                )
                if Histogram[i] > minSize:
                    maskedFile = os.path.join(
                        subsetImgsMaskedDIR, subImgBaseName + str(i) + "_masked.kea"
                    )
                else:
                    maskedFile = os.path.join(
                        subsetImgsMaskedDIR, subImgBaseName + str(i) + "_burn.kea"
                    )
                imageutils.mask_img(
                    subImage, s2BordersImage, maskedFile, "KEA", dataType, 0, 0
                )
                rastergis.pop_rat_img_stats(maskedFile, True, False)
        ratDS = None

    def performStage3SubsetsSegmentation(
        self,
        subsetImgsMaskedDIR,
        subsetSegsDIR,
        tmpDIR,
        subImgBaseName,
        segStatsInfo,
        minPxlsVal,
        distThresVal,
        bandsVal,
    ):
        imgTiles = glob.glob(
            os.path.join(subsetImgsMaskedDIR, subImgBaseName + "*_masked.kea")
        )
        for imgTile in imgTiles:
            baseName = os.path.splitext(os.path.basename(imgTile))[0]
            clumpsFile = os.path.join(subsetSegsDIR, baseName + "_segs.kea")
            kMeansCentres, imgStretchStats = self.findSegStatsFiles(
                imgTile, segStatsInfo
            )
            shepherdseg.runShepherdSegmentationPreCalcdStats(
                imgTile,
                clumpsFile,
                kMeansCentres,
                imgStretchStats,
                outputMeanImg=None,
                tmpath=os.path.join(tmpDIR, baseName + "_segstemp"),
                gdalformat="KEA",
                noStats=False,
                noStretch=False,
                noDelete=False,
                minPxls=minPxlsVal,
                distThres=distThresVal,
                bands=bandsVal,
                processInMem=False,
            )

    def mergeStage3TilesToOutput(
        self, clumpsImage, subsetSegsDIR, subsetImgsMaskedDIR, subImgBaseName
    ):
        burnTiles = glob.glob(
            os.path.join(subsetImgsMaskedDIR, subImgBaseName + "*_burn.kea")
        )
        if len(burnTiles) > 0:
            segmentation.merge_clump_images(burnTiles, clumpsImage)

        segTiles = glob.glob(os.path.join(subsetSegsDIR, subImgBaseName + "*_segs.kea"))
        segmentation.merge_clump_images(segTiles, clumpsImage)
        rastergis.pop_rat_img_stats(clumpsImage, True, True)


def perform_tiled_segmentation(
    input_img,
    clumps_img,
    tmp_dir="segtmp",
    tile_width=2000,
    tile_height=2000,
    valid_data_threshold=0.3,
    num_clusters=60,
    min_pxls=100,
    dist_thres=100,
    bands=None,
    sampling=100,
    km_max_iter=200,
):
    """
    Utility function to call the segmentation algorithm of Shepherd et al. (2019) using the tiled process outlined in Clewley et al (2015).

    :param input_img: is a string containing the name of the input file.
    :param clumps_img: is a string containing the name of the output clump file.
    :param tmpath: is a file path for intermediate files (default is to create a directory 'segtmp'). If path does current not exist then it will be created and deleted afterwards.
    :param tile_width: is an int specifying the width of the tiles used for processing (Default 2000)
    :param tile_height: is an int specifying the height of the tiles used for processing (Default 2000)
    :param valid_data_threshold: is a float (value between 0 - 1) used to specify the amount of valid image pixels (i.e., not a no data value of zero) are within a tile. Tiles failing to meet this threshold are merged with ones which do (Default 0.3).
    :param num_clusters: is an int which specifies the number of clusters within the KMeans clustering (default = 60).
    :param min_pxls: is an int which specifies the minimum number pixels within a segments (default = 100).
    :param dist_thres: specifies the distance threshold for joining the segments (default = 100, set to large number to turn off this option).
    :param bands: is an array providing a subset of image bands to use (default is None to use all bands).
    :param sampling: specify the subsampling of the image for the data used within the KMeans (default = 100; 1 == no subsampling).
    :param km_max_iter: maximum iterations for KMeans (Default 200).

    .. code:: python

        from rsgislib.segmentation import tiledsegsingle

        inputImage = 'LS5TM_20110428_sref_submask_osgb.kea'
        clumpsImage = 'LS5TM_20110428_sref_submask_osgb_clumps.kea'

        tiledsegsingle.perform_tiled_segmentation(inputImage, clumpsImage, tmpDIR='rsgislibsegtmp', tileWidth=2000, tileHeight=2000, validDataThreshold=0.3, numClusters=60, minPxls=100, distThres=100, bands=[4,5,3], sampling=100, kmMaxIter=200)

    """
    import rsgislib.tools.utils
    import rsgislib.tools.filetools

    createdTmp = False
    if not os.path.exists(tmp_dir):
        os.makedirs(tmp_dir)
        createdTmp = True
    uidStr = rsgislib.tools.utils.uid_generator()

    baseName = os.path.splitext(os.path.basename(input_img))[0] + "_" + uidStr

    tileSegInfo = os.path.join(tmp_dir, baseName + "_seginfo.json")
    segStatsDIR = os.path.join(tmp_dir, "segStats_" + uidStr)
    strchStatsBase = os.path.join(segStatsDIR, baseName + "_stch")
    kCentresBase = os.path.join(segStatsDIR, baseName + "_kcentres")
    if not os.path.exists(segStatsDIR):
        os.makedirs(segStatsDIR)

    tiledSegObj = RSGISTiledShepherdSegmentationSingleThread()

    ######################## STAGE 1 #######################
    # Stage 1 Parameters (Internal)
    stage1TileShp = os.path.join(tmp_dir, baseName + "_S1Tiles.shp")
    stage1TileRAT = os.path.join(tmp_dir, baseName + "_S1Tiles.kea")
    stage1TilesBase = baseName + "_S1Tile"
    stage1TilesImgDIR = os.path.join(tmp_dir, "s1tilesimgs_" + uidStr)
    stage1TilesMetaDIR = os.path.join(tmp_dir, "s1tilesmeta_" + uidStr)
    stage1TilesSegsDIR = os.path.join(tmp_dir, "s1tilessegs_" + uidStr)
    stage1TilesSegBordersDIR = os.path.join(tmp_dir, "s1tilessegborders_" + uidStr)
    stage1BordersImage = os.path.join(tmp_dir, baseName + "_S1Borders.kea")

    if not os.path.exists(stage1TilesImgDIR):
        os.makedirs(stage1TilesImgDIR)
    if not os.path.exists(stage1TilesSegsDIR):
        os.makedirs(stage1TilesSegsDIR)
    if not os.path.exists(stage1TilesSegBordersDIR):
        os.makedirs(stage1TilesSegBordersDIR)
    if not os.path.exists(stage1TilesMetaDIR):
        os.makedirs(stage1TilesMetaDIR)

    # Initial Tiling
    tiledSegObj.performStage1Tiling(
        input_img,
        stage1TileShp,
        stage1TileRAT,
        stage1TilesBase,
        stage1TilesMetaDIR,
        stage1TilesImgDIR,
        os.path.join(tmp_dir, "s1tilingtemp"),
        tile_width,
        tile_height,
        valid_data_threshold,
    )

    # Perform Segmentation
    tiledSegObj.performStage1TilesSegmentation(
        stage1TilesImgDIR,
        stage1TilesSegsDIR,
        tmp_dir,
        stage1TilesBase,
        tileSegInfo,
        strchStatsBase,
        kCentresBase,
        num_clusters,
        min_pxls,
        dist_thres,
        bands,
        sampling,
        km_max_iter,
    )

    # Define Boundaries
    tiledSegObj.defineStage1Boundaries(
        stage1TilesSegsDIR, stage1TilesSegBordersDIR, stage1TilesBase
    )

    # Merge the Initial Tiles
    tiledSegObj.mergeStage1TilesToOutput(
        input_img,
        stage1TilesSegsDIR,
        stage1TilesSegBordersDIR,
        stage1TilesBase,
        clumps_img,
        stage1BordersImage,
    )

    shutil.rmtree(stage1TilesImgDIR)
    shutil.rmtree(stage1TilesSegsDIR)
    shutil.rmtree(stage1TilesSegBordersDIR)
    shutil.rmtree(stage1TilesMetaDIR)
    ########################################################

    with open(tileSegInfo, "r") as f:
        jsonStrData = f.read()
    segStatsInfo = json.loads(jsonStrData)

    ######################## STAGE 2 #######################
    # Stage 2 Parameters (Internal)
    stage2TileShp = os.path.join(tmp_dir, baseName + "_S2Tiles.shp")
    stage2TileRAT = os.path.join(tmp_dir, baseName + "_S2Tiles.kea")
    stage2TilesBase = baseName + "_S2Tile"
    stage2TilesImgDIR = os.path.join(tmp_dir, "s2tilesimg_" + uidStr)
    stage2TilesMetaDIR = os.path.join(tmp_dir, "s2tilesmeta_" + uidStr)
    stage2TilesImgMaskedDIR = os.path.join(tmp_dir, "s2tilesimgmask_" + uidStr)
    stage2TilesSegsDIR = os.path.join(tmp_dir, "s2tilessegs_" + uidStr)
    stage2TilesSegBordersDIR = os.path.join(tmp_dir, "s2tilessegborders_" + uidStr)
    stage2BordersImage = os.path.join(tmp_dir, baseName + "_S2Borders.kea")

    if not os.path.exists(stage2TilesImgDIR):
        os.makedirs(stage2TilesImgDIR)
    if not os.path.exists(stage2TilesMetaDIR):
        os.makedirs(stage2TilesMetaDIR)
    if not os.path.exists(stage2TilesImgMaskedDIR):
        os.makedirs(stage2TilesImgMaskedDIR)
    if not os.path.exists(stage2TilesSegsDIR):
        os.makedirs(stage2TilesSegsDIR)
    if not os.path.exists(stage2TilesSegBordersDIR):
        os.makedirs(stage2TilesSegBordersDIR)

    # Perform offset tiling
    tiledSegObj.performStage2Tiling(
        input_img,
        stage2TileShp,
        stage2TileRAT,
        stage2TilesBase,
        stage2TilesMetaDIR,
        stage2TilesImgDIR,
        os.path.join(tmp_dir, "s2tilingtemp"),
        tile_width,
        tile_height,
        valid_data_threshold,
        stage1BordersImage,
    )

    # Perform Segmentation of the Offset Tiles
    tiledSegObj.performStage2TilesSegmentation(
        stage2TilesImgDIR,
        stage2TilesImgMaskedDIR,
        stage2TilesSegsDIR,
        stage2TilesSegBordersDIR,
        tmp_dir,
        stage2TilesBase,
        stage1BordersImage,
        segStatsInfo,
        min_pxls,
        dist_thres,
        bands,
    )

    # Merge in the next set of boundaries
    tiledSegObj.mergeStage2TilesToOutput(
        clumps_img,
        stage2TilesSegsDIR,
        stage2TilesSegBordersDIR,
        stage2TilesBase,
        stage2BordersImage,
    )

    shutil.rmtree(stage2TilesImgDIR)
    shutil.rmtree(stage2TilesMetaDIR)
    shutil.rmtree(stage2TilesImgMaskedDIR)
    shutil.rmtree(stage2TilesSegsDIR)
    shutil.rmtree(stage2TilesSegBordersDIR)
    ########################################################

    ######################## STAGE 3 #######################
    # Stage 3 Parameters (Internal)
    stage3BordersClumps = os.path.join(tmp_dir, baseName + "_S3BordersClumps.kea")
    stage3SubsetsDIR = os.path.join(tmp_dir, "s3subsetimgs_" + uidStr)
    stage3SubsetsMaskedDIR = os.path.join(tmp_dir, "s3subsetimgsmask_" + uidStr)
    stage3SubsetsSegsDIR = os.path.join(tmp_dir, "s3subsetsegs_" + uidStr)
    stage3Base = baseName + "_S3Subset"

    if not os.path.exists(stage3SubsetsDIR):
        os.makedirs(stage3SubsetsDIR)
    if not os.path.exists(stage3SubsetsMaskedDIR):
        os.makedirs(stage3SubsetsMaskedDIR)
    if not os.path.exists(stage3SubsetsSegsDIR):
        os.makedirs(stage3SubsetsSegsDIR)

    # Create the final boundary image subsets
    tiledSegObj.createStage3ImageSubsets(
        input_img,
        stage2BordersImage,
        stage3BordersClumps,
        stage3SubsetsDIR,
        stage3SubsetsMaskedDIR,
        stage3Base,
        min_pxls,
    )

    # Perform Segmentation of the stage 3 regions
    tiledSegObj.performStage3SubsetsSegmentation(
        stage3SubsetsMaskedDIR,
        stage3SubsetsSegsDIR,
        tmp_dir,
        stage3Base,
        segStatsInfo,
        min_pxls,
        dist_thres,
        bands,
    )

    # Merge the stage 3 regions into the final clumps image
    tiledSegObj.mergeStage3TilesToOutput(
        clumps_img, stage3SubsetsSegsDIR, stage3SubsetsMaskedDIR, stage3Base
    )

    shutil.rmtree(stage3SubsetsDIR)
    shutil.rmtree(stage3SubsetsMaskedDIR)
    shutil.rmtree(stage3SubsetsSegsDIR)
    ########################################################

    shutil.rmtree(segStatsDIR)
    rsgislib.tools.filetools.delete_file_with_basename(stage1BordersImage)
    rsgislib.tools.filetools.delete_file_with_basename(stage2BordersImage)
    rsgislib.tools.filetools.delete_file_with_basename(stage3BordersClumps)
    rsgislib.tools.filetools.delete_file_with_basename(stage1TileShp)
    rsgislib.tools.filetools.delete_file_with_basename(stage1TileRAT)
    rsgislib.tools.filetools.delete_file_with_basename(stage2TileShp)
    rsgislib.tools.filetools.delete_file_with_basename(stage2TileRAT)
    rsgislib.tools.filetools.delete_file_silent(tileSegInfo)
    if createdTmp:
        shutil.rmtree(tmp_dir)



#! /usr/bin/env python
############################################################################
#  classkeraspxl.py
#
#  Copyright 2020 RSGISLib.
#
#  RSGISLib: 'The remote sensing and GIS Software Library'
#
#  RSGISLib is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  RSGISLib is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with RSGISLib.  If not, see <http://www.gnu.org/licenses/>.
#
#
# Purpose:  Provide a set of functions to apply tensorflow keras classifiers
#           to remotely sensed imagery.
#
# Author: Pete Bunting
# Email: petebunting@mac.com
# Date: 14/01/2020
# Version: 1.0
#
# History:
# Version 1.0 - Created.
#
###########################################################################

from __future__ import print_function

import rsgislib
import rsgislib.rastergis

import numpy

import h5py

from osgeo import gdal

from rios import applier
from rios import cuiprogress
from rios import rat


def train_keras_pixel_classifier(
    cls_mdl, cls_info_dict, out_mdl_file=None, train_epochs=5, train_batch_size=32
):
    """
    A function which trains a neural network defined using the keras API for the classification of remotely sensed data.
    A dict of class information, as ClassInfoObj objects, is defined with the training data.

    This function requires that tensorflow and keras modules to be installed.

    :param out_mdl_file: The output model which can be loaded to perform a classification.
    :param cls_info_dict: dict (key is string with class name) of ClassInfoObj objects defining the training data.
    :param out_mdl_file: A file path to save the trained model as a hdf5 file. If None then ignored.

    """
    try:
        from keras.utils import to_categorical
    except:
        from tensorflow.keras.utils import to_categorical

    n_classes = len(cls_info_dict)
    for clsname in cls_info_dict:
        if cls_info_dict[clsname].id >= n_classes:
            raise rsgislib.RSGISPyException(
                "ClassInfoObj '{}' id ({}) is not consecutive starting from 0.".format(
                    clsname, cls_info_dict[clsname].id
                )
            )

    cls_data_dict = {}
    train_data_lst = []
    train_lbls_lst = []
    valid_data_lst = []
    valid_lbls_lst = []
    test_data_lst = []
    test_lbls_lst = []
    cls_ids = []
    for clsname in cls_info_dict:
        sgl_cls_info = {}
        print("Reading Class {} Training".format(clsname))
        f = h5py.File(cls_info_dict[clsname].train_file_h5, "r")
        sgl_cls_info["train_n_rows"] = f["DATA/DATA"].shape[0]
        sgl_cls_info["train_data"] = numpy.array(f["DATA/DATA"])
        sgl_cls_info["train_data_lbls"] = numpy.zeros(
            sgl_cls_info["train_n_rows"], dtype=int
        )
        sgl_cls_info["train_data_lbls"][...] = cls_info_dict[clsname].id
        f.close()
        train_data_lst.append(sgl_cls_info["train_data"])
        train_lbls_lst.append(sgl_cls_info["train_data_lbls"])

        print("Reading Class {} Validation".format(clsname))
        f = h5py.File(cls_info_dict[clsname].valid_file_h5, "r")
        sgl_cls_info["valid_n_rows"] = f["DATA/DATA"].shape[0]
        sgl_cls_info["valid_data"] = numpy.array(f["DATA/DATA"])
        sgl_cls_info["valid_data_lbls"] = numpy.zeros(
            sgl_cls_info["valid_n_rows"], dtype=int
        )
        sgl_cls_info["valid_data_lbls"][...] = cls_info_dict[clsname].id
        f.close()
        valid_data_lst.append(sgl_cls_info["valid_data"])
        valid_lbls_lst.append(sgl_cls_info["valid_data_lbls"])

        print("Reading Class {} Testing".format(clsname))
        f = h5py.File(cls_info_dict[clsname].test_file_h5, "r")
        sgl_cls_info["test_n_rows"] = f["DATA/DATA"].shape[0]
        sgl_cls_info["test_data"] = numpy.array(f["DATA/DATA"])
        sgl_cls_info["test_data_lbls"] = numpy.zeros(
            sgl_cls_info["test_n_rows"], dtype=int
        )
        sgl_cls_info["test_data_lbls"][...] = cls_info_dict[clsname].id
        f.close()
        test_data_lst.append(sgl_cls_info["test_data"])
        test_lbls_lst.append(sgl_cls_info["test_data_lbls"])

        cls_data_dict[clsname] = sgl_cls_info
        cls_ids.append(cls_info_dict[clsname].id)

    print("Finished Reading Data")

    train_np = numpy.concatenate(train_data_lst)
    train_lbls_np = numpy.concatenate(train_lbls_lst)
    train_lbls_keras = to_categorical(train_lbls_np, num_classes=n_classes)

    vaild_np = numpy.concatenate(valid_data_lst)
    vaild_lbls_np = numpy.concatenate(valid_lbls_lst)
    vaild_lbls_keras = to_categorical(vaild_lbls_np, num_classes=n_classes)

    test_np = numpy.concatenate(test_data_lst)
    test_lbls_np = numpy.concatenate(test_lbls_lst)
    test_lbls_keras = to_categorical(test_lbls_np, num_classes=n_classes)

    print("Start Training Model")
    cls_mdl.fit(
        train_np,
        train_lbls_keras,
        epochs=train_epochs,
        batch_size=train_batch_size,
        validation_data=(vaild_np, vaild_lbls_keras),
    )
    print("Finished Training Model")
    cls_mdl.summary()

    loss_and_metrics = cls_mdl.evaluate(
        test_np, test_lbls_keras, batch_size=train_batch_size
    )
    eval_metric_names = cls_mdl.metrics_names
    for eval_name, eval_val in zip(eval_metric_names, loss_and_metrics):
        print("{} = {}".format(eval_name, eval_val))

    if out_mdl_file is not None:
        cls_mdl.save(out_mdl_file)


def apply_keras_pixel_classifier(
    class_train_info,
    keras_cls_mdl,
    in_img_mask,
    img_mask_val,
    img_file_info,
    out_class_img,
    gdalformat,
    pred_batch_size=32,
    class_clr_names=True,
):
    """
    This function applies a trained single pixel keras model to an image. The function train_keras_pixel_classifier
    can be used to train such as model. The output image will contain the hard membership of the predicted class.

    :param class_train_info: dict (where the key is the class name) of rsgislib.classification.ClassInfoObj
                           objects which will be used to train the classifier (i.e., train_keras_pixel_classifier()),
                           provide pixel value id and RGB class values.
    :param keras_cls_mdl: a trained keras model object, with a input dimensions equivlent to the number of image
                          bands specified in the imgFileInfo input and output layer which provides an output array
                          of the length of the number of classes.
    :param in_img_mask: is an image file providing a mask to specify where should be classified. Simplest mask is all the
                    valid data regions (rsgislib.imageutils.gen_valid_mask)
    :param img_mask_val: the pixel value within the imgMask to limit the region to which the classification is applied.
                       Can be used to create a heirachical classification.
    :param img_file_info: a list of rsgislib.imageutils.ImageBandInfo objects (also used within
                        rsgislib.zonalstats.extract_zone_img_band_values_to_hdf) to identify which images and bands are to
                        be used for the classification so it adheres to the training data.
    :param out_class_img: Output image which will contain the hard classification.
    :param gdalformat: is the output image format - all GDAL supported formats are supported.
    :param pred_batch_size: the batch size used for the classification.
    :param class_clr_names: default is True and therefore a colour table will the colours specified in ClassInfoObj
                          and a class_names (from classTrainInfo) column will be added to the output file.

    """

    def _applyKerasPxlClassifier(info, inputs, outputs, otherargs):
        outClassIdVals = numpy.zeros_like(inputs.imageMask, dtype=numpy.uint16)
        if numpy.any(inputs.imageMask == otherargs.mskVal):
            n_pxls = inputs.imageMask.shape[1] * inputs.imageMask.shape[2]
            outClassIdVals = outClassIdVals.flatten()
            imgMaskVals = inputs.imageMask.flatten()
            classVars = numpy.zeros(
                (n_pxls, otherargs.numClassVars), dtype=numpy.float32
            )
            # Array index which can be used to populate the output array following masking etc.
            ID = numpy.arange(imgMaskVals.shape[0])
            classVarsIdx = 0
            for imgFile in otherargs.imgFileInfo:
                imgArr = inputs.__dict__[imgFile.name]
                for band in imgFile.bands:
                    classVars[..., classVarsIdx] = imgArr[(band - 1)].flatten()
                    classVarsIdx = classVarsIdx + 1
            classVars = classVars[imgMaskVals == otherargs.mskVal]
            ID = ID[imgMaskVals == otherargs.mskVal]
            preds_idxs = numpy.argmax(
                otherargs.classifier.predict(
                    classVars, batch_size=otherargs.pred_batch_size
                ),
                axis=1,
            )
            preds_cls_ids = numpy.zeros_like(preds_idxs, dtype=numpy.uint16)
            for cld_id, idx in zip(
                otherargs.cls_id_lut, numpy.arange(0, len(otherargs.cls_id_lut))
            ):
                preds_cls_ids[preds_idxs == idx] = cld_id

            outClassIdVals[ID] = preds_cls_ids
            outClassIdVals = numpy.expand_dims(
                outClassIdVals.reshape(
                    (inputs.imageMask.shape[1], inputs.imageMask.shape[2])
                ),
                axis=0,
            )
        outputs.outclsimage = outClassIdVals

    infiles = applier.FilenameAssociations()
    infiles.imageMask = in_img_mask
    numClassVars = 0
    for imgFile in img_file_info:
        infiles.__dict__[imgFile.name] = imgFile.fileName
        numClassVars = numClassVars + len(imgFile.bands)

    n_classes = len(class_train_info)
    cls_id_lut = numpy.zeros(n_classes)
    for clsname in class_train_info:
        if class_train_info[clsname].id >= n_classes:
            raise rsgislib.RSGISPyException(
                "ClassInfoObj '{}' id ({}) is not consecutive starting from 0.".format(
                    clsname, class_train_info[clsname].id
                )
            )
        cls_id_lut[class_train_info[clsname].id] = class_train_info[clsname].out_id

    outfiles = applier.FilenameAssociations()
    outfiles.outclsimage = out_class_img
    otherargs = applier.OtherInputs()
    otherargs.classifier = keras_cls_mdl
    otherargs.pred_batch_size = pred_batch_size
    otherargs.mskVal = img_mask_val
    otherargs.numClassVars = numClassVars
    otherargs.imgFileInfo = img_file_info
    otherargs.n_classes = n_classes
    otherargs.cls_id_lut = cls_id_lut

    try:
        import tqdm

        progress_bar = rsgislib.TQDMProgressBar()
    except:
        progress_bar = cuiprogress.GDALProgressBar()

    aControls = applier.ApplierControls()
    aControls.progress = progress_bar
    aControls.drivername = gdalformat
    aControls.omitPyramids = True
    aControls.calcStats = False
    print("Applying the Classifier")
    applier.apply(
        _applyKerasPxlClassifier, infiles, outfiles, otherargs, controls=aControls
    )
    print("Completed Classification")

    if class_clr_names:
        rsgislib.rastergis.pop_rat_img_stats(
            out_class_img, add_clr_tab=True, calc_pyramids=True, ignore_zero=True
        )
        ratDataset = gdal.Open(out_class_img, gdal.GA_Update)
        red = rat.readColumn(ratDataset, "Red")
        green = rat.readColumn(ratDataset, "Green")
        blue = rat.readColumn(ratDataset, "Blue")
        class_names = numpy.empty_like(red, dtype=numpy.dtype("a255"))
        class_names[...] = ""

        for classKey in class_train_info:
            print("Apply Colour to class '" + classKey + "'")
            red[class_train_info[classKey].out_id] = class_train_info[classKey].red
            green[class_train_info[classKey].out_id] = class_train_info[classKey].green
            blue[class_train_info[classKey].out_id] = class_train_info[classKey].blue
            class_names[class_train_info[classKey].out_id] = classKey

        rat.writeColumn(ratDataset, "Red", red)
        rat.writeColumn(ratDataset, "Green", green)
        rat.writeColumn(ratDataset, "Blue", blue)
        rat.writeColumn(ratDataset, "class_names", class_names)
        ratDataset = None
